import os

from cffi import FFI

#TODO: configure this better
htslib_dir=os.path.join('submodules', 'samtools-1.3.1', 'htslib-1.3.1')

libraries=['z', 'lzma', 'bz2', 'pthread']
library_dirs=[htslib_dir]
src_dir='src'

ffibuilder = FFI()
ffibuilder.set_source("libmedaka",
    r"""
    #include "medaka_counts.h"
    """,
    libraries=libraries,
    library_dirs=library_dirs,
    include_dirs=[src_dir, htslib_dir],
    sources=[os.path.join(src_dir, 'medaka_counts.c')],
    extra_compile_args=['-std=c99', '-msse3', '-O3'],
    extra_objects=['libhts.a']
)

with open(os.path.join(src_dir, 'medaka_counts.h'), 'r') as fh:
    ffibuilder.cdef(fh.read())


if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



import logging
from itertools import chain

import numpy as np

from medaka.common import (_gap_, Region, get_sample_overlap, get_sample_index_from_files,
                           load_yaml_data, yield_from_feature_files, _label_decod_path_)


def write_fasta(filename, contigs):
    with open(filename, 'w') as fasta:
        for name, seq in contigs:
            fasta.write('>{}\n{}\n'.format(name, seq))


def stitch_from_probs(probs_hdfs, regions=None, model_yml=None):
    """Decode and join overlapping label probabilities from hdf to assemble complete sequence

    :param probs_hdfs: iterable of hdf filepaths.
    :ref_names: iterable of ref_names to limit stitching to.
    :returns: list of (contig_name, sequence)

    """
    logger = logging.getLogger(__package__)
    logger.name = 'Stitch'

    label_decoding = load_yaml_data(probs_hdfs[0], _label_decod_path_)
    if label_decoding is None:
        if model_yml is not None:
            logging.info("Loading label encoding from {}".format(model_yml))
            label_decoding = load_yaml_data(model_yml, _label_decod_path_)
        else:
            raise ValueError('Cannot decode probabilities without label decoding')

    logger.debug("Label decoding is:\n{}".format('\n'.join(
        '{}: {}'.format(i, x) for i, x in enumerate(label_decoding)
    )))
    index = get_sample_index_from_files(probs_hdfs, 'hdf')
    if regions is None:
        ref_names = index.keys()
    else:
        #TODO: respect entire region specification
        ref_names = list()
        for region in (Region.from_string(r) for r in regions):
            if region.start is None or region.end is None:
                logger.warning("Ignoring start:end for '{}'.".format(region))
            ref_names.append(region.ref_name)

    get_pos = lambda s, i: '{}.{}'.format(s.positions[i]['major'] + 1, s.positions[i]['minor'])
    ref_assemblies = []
    for ref_name in ref_names:
        logger.info("Processing {}.".format(ref_name))
        data_gen = yield_from_feature_files(probs_hdfs, ref_names=(ref_name,), index=index)
        seq=''
        s1 = next(data_gen)
        start = get_pos(s1, 0)
        start_1_ind = None
        for s2 in chain(data_gen, (None,)):
            if s2 is None:  # s1 is last chunk
                end_1_ind = None
            else:
                end_1_ind, start_2_ind = get_sample_overlap(s1, s2)

            best = np.argmax(s1.label_probs[start_1_ind:end_1_ind], -1)
            seq += ''.join([label_decoding[x] for x in best]).replace(_gap_, '')
            if end_1_ind is None:
                key = '{}:{}-{}'.format(s1.ref_name, start, get_pos(s1, -1))
                ref_assemblies.append((key, seq))
                seq = ''
                if start_2_ind is None:
                    msg = 'There is no overlap betwen {} and {}'
                    logging.info(msg.format(s1.name, s2.name))
                    start = get_pos(s2, 0)

            s1 = s2
            start_1_ind = start_2_ind
    return ref_assemblies


def stitch(args):
    joined = stitch_from_probs(args.inputs, regions=args.regions, model_yml=args.model_yml)
    write_fasta(args.output, joined)



import itertools
import logging
import numpy as np
import pysam
from copy import copy
from operator import attrgetter
from medaka.common import _gap_, encoding
from medaka.common import get_pairs, yield_compressed_pairs, get_pairs_with_hp_len


class TruthAlignment(object):
    def __init__(self, alignment):
        """Create a TruthAlignment oblist from a `pysam.libcalignedsegment.AlignedSegment` object.

        :param alignment: `pysam.libcalignedsegment.AlignedSegment` object.
        """
        self.aln = alignment  # so we can get positions and labels later
        # initialise start and end (which might be moved)
        self.start = self.aln.reference_start  # zero-based
        self.end = self.aln.reference_end
        self.is_kept = True
        self.logger = logging.getLogger('TruthAlign')

    def get_overlap_with(self, other):
        first, second = sorted((self, other), key=attrgetter('aln.reference_start'))
        overlap = first.aln.reference_end > second.aln.reference_start
        if overlap:
            # return positions of start and end of overlapping region
            return second.aln.reference_start, first.aln.reference_end
        else:
            return None

    @staticmethod
    def filter_alignments(alignments, min_length=1000, length_ratio=2.0, overlap_fraction=0.5,
                         start=0, end=None):
        """Filter alignments to yield only segments suitable for training.

        :param alignments: iterable of `TruthAlignment` objects.
        :param min_length: int, minimum length of alignment segment.
        :param length_ratio: float, in cases of overlap, ratio of longer segment / shorter segment
            above which the longer segment is assumed to be more reliable.
        :param overlap_fraction: float, length_ratio: float, in cases of overlap, fraction of
            shorter segment overlapping with the longer segment above which the segments are
            considered highly overlapping.
        :param start: zero-based start position in reference coordinates. Any alignments which
            start before this value will take this new starting coordinate.
        :param end: zero-based end position in reference coordinates. and alignments
            which end after this value will take this new ending coordinate.
        :returns: list of `TruthAlignment` objects

        """
        filtered_alignments = [copy(a) for a in alignments]  # don't want to modify original alignments
        for al_i, al_j in itertools.combinations(filtered_alignments, 2):
            first, second = sorted((al_i, al_j), key=attrgetter('aln.reference_start'))
            overlap = first.get_overlap_with(second)
            if overlap is None:
                continue
            ovlp_start, ovlp_end = overlap
            shorter, longer = sorted((al_i, al_j), key=attrgetter('aln.reference_length'))
            length_ratio_ij = longer.aln.reference_length / shorter.aln.reference_length
            overlap_fraction_ij = (ovlp_end - ovlp_start) / shorter.aln.reference_length
            # 4 cases
            if length_ratio_ij < length_ratio:  # we don't trust one more than the other
                if overlap_fraction_ij >= overlap_fraction:
                    # 1) they overlap a lot; we have significant ambiguity, discard both
                    shorter.is_kept = False
                    longer.is_kept = False
                else:
                    # 2) they overlap a little; just trim overlapping portions of both alignments
                    first.end = ovlp_start
                    second.start = ovlp_end
            else:  # we trust the longer one more than the shorter one
                if overlap_fraction_ij >= overlap_fraction:
                    # 3) they overlap a lot; discard shorter alignment
                    shorter.is_kept = False
                else:
                    # 4) they overlap a little; trim overlapping portion of shorter alignment
                    second.start = ovlp_end

        # trim starts and ends if needed
        if start > 0 or end is not None:
            for al in filtered_alignments:
                if start > 0:
                    al.start = max(start, al.start)
                if end is not None:
                    al.end = min(end, al.end)
        # do filtering
        filtered_alignments = [al for al in filtered_alignments
                               if (al.is_kept and al.end - al.start >= min_length)]
        filtered_alignments.sort(key=attrgetter('start'))
        return filtered_alignments

    @staticmethod
    def bam_to_alignments(truth_bam, ref_name, start=None, end=None):
        """Create list of TruthAlignment objects from a bam of Truth aligned to ref.

        :param truth_bam: (sorted indexed) bam with true sequence aligned to reference
        :param ref: name of reference to process
        :param start: starting position within reference
        :param end: ending position within reference
            (all alignments with any overlap with the interval start:end will be retrieved)
        :returns: tuple(positions, encoded_label_array)

            - positions: numpy structured array with 'ref_major'
              (reference position index) and 'ref_minor'
              (trailing insertion index) fields.

            - feature_array: 1D numpy array of encoded labels
        """
        with pysam.AlignmentFile(truth_bam, 'rb') as bamfile:
            aln_reads = bamfile.fetch(reference=ref_name, start=start, end=end)
            alignments = [TruthAlignment(r) for r in aln_reads if not (r.is_unmapped or r.is_secondary)]
            alignments.sort(key=attrgetter('start'))
        logger = logging.getLogger("TruthAlign")
        logger.info("Retrieved {} alignments.".format(len(alignments)))
        return alignments

    def get_positions_and_labels(self, start=None, end=None, ref_compr_rle=None, mock_compr=False, is_compressed=False, rle_dtype=False):
        """Create labels and positions array.

        :param start: starting position within reference
        :param end: ending position within reference
        :param ref_compr_rle: np.ndarray, dtype [('start', int), ('end', int)]
            containing uncompressed start position and length for each compressed position.
        :param mock_compr: bool, labels will be encoded as for a compressed sequence, even if ref_compr_rle is None.
        :param is_compressed: bool, whether the sequence has been run length encoded.
        :param rle_dtype: bool, whether to force dtype of labels to be (int base , int length).
        :returns: tuple(positions, encoded_label_array)

            - positions: numpy structured array with 'ref_major'
              (reference position index) and 'ref_minor'
              (trailing insertion index) fields.

            - label_array: 1D numpy array of labels
        """
        if start is None:
            start = 0
        if end is None:
            end = float('Inf')
        # ensure start and end fall within filtered region
        start = max(start, self.start)
        end = min(end, self.end)

        positions_labels = []
        if is_compressed:
            pairs = yield_compressed_pairs(self.aln, ref_compr_rle)
        elif mock_compr:
            pairs = get_pairs_with_hp_len(self.aln, ref_compr_rle)
        else:
            pairs = get_pairs(self.aln)

        ins_count = 0

        for pair in itertools.dropwhile(lambda x: (x.rpos is None)
                                        or (x.rpos < start), pairs):
            if ((pair.rpos == self.aln.reference_end ) or
                (pair.rpos is not None and pair.rpos >= end)):
                break
            if pair.rpos is None:
                ins_count += 1
            else:
                ins_count = 0
                current_pos = pair.rpos
            pos = (current_pos, ins_count)
            label = pair.qbase.upper() if pair.qbase else _gap_
            if label == 'N':
                self.logger.info('Found {} at pos {}'.format(label, pos))
            label = encoding[label]
            if is_compressed or mock_compr:
                label = (label, pair.qlen)
            elif rle_dtype:
                label = (label, 1)
            positions_labels.append((pos, label))

        if is_compressed or mock_compr or rle_dtype:
            dtype = [('base', int), ('run_length', int)]
        else:
            dtype = int
        label_array = np.empty(shape=len(positions_labels), dtype=dtype)
        positions = np.empty(len(positions_labels),
                             dtype=[('major', int), ('minor', int)])

        for i, (pos, label) in enumerate(positions_labels):
            positions[i] = pos
            label_array[i] = label

        return positions, label_array



from medaka.features import FeatureEncoder, pileup_counts
from medaka.common import Region
import sys

import numpy as np

import argparse
from timeit import default_timer as now
import logging

def main():
    # Entry point for testing/checking
    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    np.set_printoptions(precision=4, linewidth=100)
    
    
    parser = argparse.ArgumentParser('medaka', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('bam', help='alignment file.')
    parser.add_argument('region', help='alignment region to sample.')
    parser.add_argument('--print', action='store_true', help='print counts.')
    parser.add_argument('--dtypes', nargs='+', help='perform a multi-datatype tests.')
    args = parser.parse_args()
    
    region = Region.from_string(args.region)
    
    kwargs={
        'log_min': None,
        'max_hp_len': 1,
        'is_compressed': False,
        'consensus_as_ref': False,
        'ref_mode': None,
        'with_depth': False,
    }
    
    def _print(samples):
       if args.print:
           for p, f in zip(samples.positions, samples.features):
               print('{}\t{}\t0\t{}\t{}'.format(p[0], p[1], '\t'.join('{:.3f}'.format(x) if x>0.0 else '-' for x in f), sum(f)))
    
    dtype_options = [('',)]
    if args.dtypes is not None:
        dtype_options.append(args.dtypes)
    norm_options = (None, 'total', 'fwd_rev')
    
    for dtypes in dtype_options:
        kwargs['dtypes'] = dtypes
        for norm in norm_options:
            kwargs['normalise'] = norm
    
            print("###########################################################")
            print(kwargs)
            encoder = FeatureEncoder(**kwargs)
        
            # py-style
            t0=now()
            samples = encoder.bam_to_sample(args.bam, region, force_py=True)
            t1=now()
            if not samples.is_empty:
                print(samples.features.shape)
                _print(samples)
            else:
                print("Samples is empty")
            print("---------------------")
        
            # C-style
            t2=now()
            samples = encoder.bam_to_sample(args.bam, region)
            t3=now()
            if not samples.is_empty:
                print(samples.features.shape)
                _print(samples)
            else:
                print("Samples is empty")
        
            print("pysam time:", t1 - t0)
            print("hts time:", t3 - t2)


if __name__ == '__main__':
    main()




from collections import OrderedDict, namedtuple, defaultdict, Counter
from concurrent.futures import ProcessPoolExecutor
from copy import copy
import errno
import functools
import itertools
import logging
import os
import logging
from pkg_resources import resource_filename
import queue
import re
import threading
import time
from timeit import default_timer as now
import yaml

from Bio import SeqIO
import h5py
import numpy as np
import pysam

# don't import keras here, as it means a slow import of tensorflow
#from keras.utils.np_utils import to_categorical


# Codec for converting tview output to ints.
#TODO: this can likely be renomved
_gap_ = '*'
_ref_gap_ = '#'
_read_sep_ = ' '
_alphabet_ = 'ACGT'
_extra_bases_ = 'N'
#TODO: change name of these
decoding = _gap_ + _alphabet_.lower() + _alphabet_.upper() + _read_sep_ + _extra_bases_
# store encoding in ordered dict as the order will always be the same
# (we need order to be the same for training and inference)
encoding = OrderedDict(((a, i) for i, a in enumerate(decoding)))

AlignPos = namedtuple('AlignPos', ('qpos', 'qbase', 'rpos', 'rbase'))
ComprAlignPos = namedtuple('AlignPos', ('qpos', 'qbase', 'qlen', 'rpos', 'rbase', 'rlen'))

_feature_opt_path_ = 'medaka_features_kwargs'
_model_opt_path_ = 'medaka_model_kwargs'
_sample_path_ = 'samples'
_label_decod_path_ = 'medaka_label_decoding'
_feature_decoding_path_ = 'medaka_feature_decoding'
_label_counts_path_ = 'medaka_label_counts'
_feature_batches_path_ = 'medaka_feature_batches'
_label_batches_path_ = 'medaka_label_batches'


#TODO: refactor all this..
# We might consider creating a Sample class with methods to encode, decode sample name
# and perhaps to write/read Samples to/from hdf etc
_Sample = namedtuple('Sample', ['ref_name', 'features', 'labels', 'ref_seq', 'positions', 'label_probs'])
class Sample(_Sample):

    def _get_pos(self, index):
        p = self.positions
        return p['major'][index], p['minor'][index]

    @property
    def first_pos(self):
        """Zero-based first reference co-ordinate."""
        return self._get_pos(0)

    @property
    def last_pos(self):
        """Zero-based (end inclusive) last reference co-ordinate."""
        return self._get_pos(-1)

    @property
    def span(self):
        """Size of sample in terms of reference positions."""
        return self.last_pos[0] - self.first_pos[0]

    @property
    def is_empty(self):
        return self.size == 0

    @property
    def size(self):
        return len(self.positions)

    @property
    def name(self):
        """Create zero-based (end inclusive) samtools-style region string."""
        fmaj, fmin = self.first_pos
        lmaj, lmin = self.last_pos
        return '{}:{}.{}-{}.{}'.format(
            self.ref_name, fmaj, fmin, lmaj + 1, lmin)


    @staticmethod
    def decode_sample_name(name):
        """Decode a the result of Sample.name into a dict.
    
        :param key: `Sample` object.
        :returns: dict.
        """
        d = None
        name_decoder = re.compile(r"(?P<ref_name>.+):(?P<start>\d+\.\d+)-(?P<end>\d+\.\d+)")
        m = re.match(name_decoder, name)
        if m is not None:
            d = m.groupdict()
        return d

    def chunks(self, chunk_len=1000, overlap=200):
        """Create overlapping chunks of self.
        
        :param chunk_len: chunk length (number of columns)
        :param overlap: overlap length.

        :yields: chunked :py:class:`Sample` instances.
    
        """
        chunker = functools.partial(sliding_window,
            window=chunk_len, step=chunk_len - overlap, axis=0)
        sample = self._asdict()
        chunkers = {
            field: chunker(sample[field])
            if sample[field] is not None else itertools.repeat(None)
            for field in sample.keys()
        }

        for i, pos in enumerate(chunkers['positions']):
            fields = set(sample.keys()) - set(['positions', 'ref_name'])
            new_sample = {
                'positions':pos, 'ref_name':sample['ref_name']
            }
            for field in fields:
                new_sample[field] = next(chunkers[field])
            yield Sample(**new_sample)


#TODO: refactor this
_Region = namedtuple('Region', 'ref_name start end')
class Region(_Region):

    @property
    def name(self):
        """Samtools-style region string, zero-base end exclusive."""
        return self.__str__()


    def __str__(self):
        # This will be zero-based, end exclusive
        return '{}:{}-{}'.format(self.ref_name, self.start, self.end)


    @property
    def size(self):
        return self.end - self.start


    @classmethod
    def from_string(cls, region):
        """Parse region strings into `Region` objects.
    
        :param regions: iterable of str
    
        >>> parse_regions(['Ecoli'])[0]
        Region(ref_name='Ecoli', start=None, end=None)
        >>> parse_regions(['Ecoli:1000-2000'])[0]
        Region(ref_name='Ecoli', start=1000, end=2000)
        >>> parse_regions(['Ecoli:1000'])[0]
        Region(ref_name='Ecoli', start=0, end=1000)
        >>> parse_regions(['Ecoli:500-'])[0]
        Region(ref_name='Ecoli', start=500, end=None)
    
        """
        if ':' not in region:
            ref_name, start, end = region, None, None
        else:
            start, end = None, None
            ref_name, bounds = region.split(':')
            if bounds[0] == '-':
                start = 0
                end = int(bounds[1:])
            elif bounds[-1] == '-':
                start = int(bounds[:-1])
                end = None
            else:
                start, end = [int(b) for b in bounds.split('-')]
        return cls(ref_name, start, end)


    def split(region, size, overlap=0):
        """Split region into sub-regions.

        :param size: size of sub-regions.
        :param overlap: overlap between ends of sub-regions.

        :returns: a list of sub-regions.

        """
        regions = [
            Region(region.ref_name, start, stop) for (start, stop) in
            segment_limits(region.start, region.end, segment_len=size, overlap_len=overlap)
        ]
        # correct end co-ordinate of the last
        last = regions[-1]
        regions[-1] = Region(last.ref_name, last.start, last.end + 1)
        return regions


def get_regions(bam, region_strs=None):
    """Create `Region` objects from a bam and region strings.

    :param bam: `.bam` file.
    :param region_strs: iterable of str in zero-based (samtools-like)
        region format e.g. ref:start-end or filepath containing a
        region string per line.

    :returns: list of `Region` objects.
    """
    with pysam.AlignmentFile(bam) as bam_fh:
        ref_lengths = dict(zip(bam_fh.references, bam_fh.lengths))
    if region_strs is not None:
        if os.path.isfile(region_strs[0]):
            with open(region_strs[0]) as fh:
                region_strs = [l.strip() for l in fh.readlines()]

        regions = []
        for r in (Region.from_string(x) for x in region_strs):
            start = r.start if r.start is not None else 0
            end = r.end if r.end is not None else ref_lengths[r.ref_name]
            regions.append(Region(r.ref_name, start, end))
    else:
        regions = [Region(ref_name, 0, end) for ref_name, end in ref_lengths.items()]

    return regions


#TODO refactor the below two function into single interface
def write_sample_to_hdf(s, hdf_fh):
    """Write a sample to HDF5.

    :param s: `Sample` object.
    :param hdf_fh: `h5py.File` object.
    """
    grp = s.name
    for field in s._fields:
        if getattr(s, field) is not None:
            data = getattr(s, field)
            if isinstance(data, np.ndarray) and isinstance(data[0], np.unicode):
                data = np.char.encode(data)
            hdf_fh['{}/{}/{}'.format(_sample_path_, grp, field)] = data


def write_samples_to_hdf(fname, samples):
    """Write samples to hdf, ensuring a sample is not written twice and maintaining
    a count of labels seen.

    :param fname: str, output filepath.
    :param samples: iterable of `Sample` objects.
    """
    logging.info("Writing samples to {}".format(fname))
    samples_seen = set()
    labels_counter = Counter()
    with h5py.File(fname, 'w') as hdf:
        for s in samples:
            s_name = s.name
            logging.debug("Written sample {}".format(s_name))
            if s_name not in samples_seen:
                write_sample_to_hdf(s, hdf)
            else:
                logging.debug('Not writing {} as it is present already'.format(s_name))
            samples_seen.add(s_name)
            if s.labels is not None:
                if len(s.labels.dtype) == 2:
                    labels_counter.update([tuple(l) for l in s.labels])
                else:
                    labels_counter.update(s.labels)
        hdf[_label_counts_path_] = yaml.dump(labels_counter)

    h = lambda l: (decoding[l[0]], l[1]) if type(l) == tuple else l
    logging.info("Label counts:\n{}".format('\n'.join(
        ['{}: {}'.format(h(label), count) for label, count in labels_counter.items()]
    )))


def lengths_to_rle(lengths):
    runs = np.empty(len(lengths), dtype=[('start', int), ('length', int)])
    runs['length'][:] = lengths
    runs['start'][0] = 0
    runs['start'][1:] = np.cumsum(lengths)[:-1]
    return runs


def get_pairs(aln):
    """Return generator of pairs.

    :param aln: `pysam.AlignedSegment` object.
    :returns: generator of `ComprAlignPos` objects.
    """
    seq = aln.query_sequence
    pairs = (ComprAlignPos(qp, seq[qp], 1, rp, rb, 1) if qp is not None
            else ComprAlignPos(qp, None, 1, rp, rb, 1)
            for qp, rp, rb in aln.get_aligned_pairs(with_seq=True))

    return pairs


def seq_to_hp_lens(seq):
    """Return array of HP lengths for every position in the sequence

    :param seq: str, sequence
    :returns: `np.ndarray` containing HP lengths (the same length as seq).
    """
    q_rle = rle(np.fromiter(seq, dtype='U1', count=len(seq)), low_mem=True)
    # get rle encoding for every position (so we can slice with qb instead of
    # searching)
    qlens = np.repeat(q_rle['length'], q_rle['length'])
    return qlens


def get_pairs_with_hp_len(aln, ref_seq):
    """Return generator of pairs in which the qbase is not just the base,
    but is decorated with extra information. E.g. an A which is part of a basecalled
    6mer HP would be AAAAAA.

    :param aln: `pysam.AlignedSegment` object.
    :param ref_seq: str containing reference sequence or result of
        `seq_to_hp_lens` (ref_seq).

    :yields: `ComprAlignPos` objects.
    """
    seq = aln.query_sequence
    qlens = seq_to_hp_lens(seq)

    rlens = seq_to_hp_lens(ref_seq) if isinstance(ref_seq, str) else ref_seq

    h = lambda ar, i, alt: ar[i] if i is not None else alt
    for qp, rp, rb in aln.get_aligned_pairs(with_seq=True):
        a = ComprAlignPos(qpos=qp, qbase=h(seq, qp, None), qlen=h(qlens, qp, 1),
                          rpos=rp, rbase=rb, rlen=h(rlens, rp, 1))
        yield a


def yield_compressed_pairs(aln, ref_rle):
    """Yield ComprAlignPos objects for aligned pairs of an AlignedSegment.

    :param aln: pysam AlignedSegment.
    :param ref_rle: structured array with columns 'start' and 'length' encoding ref.
    :yields: `ComprAlignPos` objects.
    """
    seq = aln.query_sequence
    qlens = aln.query_qualities
    if qlens is None:
        raise ValueError('Found an alignment without qscores, try filter to only primary alignments.')
    qrle = lengths_to_rle(qlens)
    h = lambda ar, i, alt: ar[i] if i is not None else alt
    for qp, rp, rb in aln.get_aligned_pairs(with_seq=True):
        a = ComprAlignPos(qpos=qp, qbase=h(seq, qp, None), qlen=h(qlens, qp, 1),
                          rpos=rp, rbase=rb, rlen=h(ref_rle['length'], rp, 1))
        yield a


def load_yaml_data(fname, group):
    """Load a yml str either from hdf or .yml file"""
    data = None
    if os.path.splitext(fname)[-1] == '.yml':
        with open(fname) as fh:
            yml_str = fh.read()
        d = yaml.load(yml_str)
        if group in d:
            data = d[group]
    else:
        with h5py.File(fname, 'r') as hdf:
            if group in hdf:
                yml_str = hdf[group][()]
                data = yaml.load(yml_str)
    return data


def write_yaml_data(fname, data):
    """Save a data structure to a yml str either within a hdf or .yml file"""
    if os.path.splitext(fname)[-1] == '.yml':
        with open(fname, 'w') as fh:
            fh.write(yaml.dump(data))
    else:
        with h5py.File(fname) as hdf:
            for group, d in data.items():
                if group in hdf:
                    del hdf[group]
                hdf[group] = yaml.dump(d)


def load_sample_from_hdf(key, hdf_fh):
    """Load `Sample` object from HDF5

    :param key: str, sample name.
    :param hdf_fh: `h5.File` object.
    :returns: `Sample` object.
    """
    s = {}
    for field in Sample._fields:
        pth = '{}/{}/{}'.format(_sample_path_, key, field)
        if pth in hdf_fh:
            s[field] = hdf_fh[pth][()]
            if isinstance(s[field], np.ndarray) and isinstance(s[field][0], type(b'')):
                s[field] = np.char.decode(s[field])
        else:
            s[field] = None
    return Sample(**s)


def yield_from_feature_files(fnames, ref_names=None, index=None, samples=None):
    """Yield `Sample` objects from one or more feature files.

    :param fnames: iterable of str of filepaths.
    :ref_names: iterable of str, only process these references.
    :index: result of `get_sample_index_from_files`
    :samples: iterable of sample names to yield (in order in which they are supplied).
    :yields: `Sample` objects.
    """
    handles = {fname: h5py.File(fname, 'r') for fname in fnames}

    if samples is not None:
        # yield samples in the order they are asked for
        for sample, fname in samples:
            yield load_sample_from_hdf(sample, handles[fname])
    else:
        # yield samples sorted by ref_name and start
        if index is None:
            index = get_sample_index_from_files(fnames, 'hdf')
        if ref_names is None:
            ref_names = sorted(index.keys())
        for ref_name in ref_names:
            for d in index[ref_name]:
                yield load_sample_from_hdf(d['key'], handles[d['filename']])

    for fh in handles.values():
        fh.close()


def load_from_feature_files(fnames, key_fnames):
    """Yield `Sample` objects in order from the result of `save_feature_file`.

    :param fnames: iterable of str of filepaths.
    :key_fnames: iterable of (str sample key, fname) tuples.
    :returns: list of `Sample` objects.
    """
    t0 = now()
    handles = { fname: h5py.File(fname, 'r') for fname in fnames}
    t1 = now()
    logging.info('Creating h5py.File objs took {}s'.format(t1 - t0))

    t0 = now()
    samples = [ load_sample_from_hdf(key, handles[fname]) for (key, fname) in key_fnames]
    t1 = now()
    logging.info('Loading samples took {}s'.format(t1 - t0))
    for fh in handles.values():
        fh.close()
    return samples


def load_feature_file(fname, ref_names=None):
    """Load `Sample` objects from HDF5.

    :param fname: str, HDF5 filepath.
    :param ref_names: iterable of str, only fetch samples for this reference.
    :returns: tuple of `Sample` objects.
    """
    return tuple([s for s in yield_from_feature_files([fname], ref_names)])


def get_sample_overlap(s1, s2):
    """Calculate overlap (in pileup columns) between two `Sample` objects.

    :param s1: First `Sample` object.
    :param s2: Second `Sample` object.
    :returns: (int end1, int start2) such that s1.positions[:end1] could be
        concatenated with s2.positions[start2:] to join them without gaps or overlap.
    """
    ovl_start_ind1 = np.searchsorted(s1.positions, s2.positions[0])
    if ovl_start_ind1 == len(s1.positions):
        # they don't overlap
        print('{} and {} do not overlap'.format(s1.name, s2.name))
        return None, None

    ovl_end_ind2 = np.searchsorted(s2.positions, s1.positions[-1], side='right')
    pos1_ovl = s1.positions[ovl_start_ind1:]
    pos2_ovl = s2.positions[0:ovl_end_ind2]
    assert len(pos1_ovl) == len(pos2_ovl)
    overlap_len = len(pos1_ovl)
    pad_1 = overlap_len // 2
    pad_2 = overlap_len - pad_1
    end_1_ind = ovl_start_ind1 + pad_1
    start_2_ind = ovl_end_ind2 - pad_2

    contr_1 = s1.positions[ovl_start_ind1:end_1_ind]
    contr_2 = s2.positions[start_2_ind:ovl_end_ind2]
    assert len(contr_1) + len(contr_2) == overlap_len

    return end_1_ind, start_2_ind


def get_sample_index_from_files(fnames, filetype='hdf', max_samples=np.inf):
    """Create index of samples from one or more HDF5 or fasta files.

    :param fnames: iterable of str of filepaths.
    :param filetype: str, 'hdf' or 'fasta'.
    :param max_samples: int, stop after max_samples.
    :returns: dict of lists, indexed by reference.
        each list contains a list of samples dicts, and is sorted by sample start.
    """

    count = 0
    ref_names = defaultdict(list)
    if filetype == 'hdf':
        fhs = (h5py.File(f) for f in fnames)
    elif filetype == 'fasta':
        fhs = (SeqIO.index(f, 'fasta') for f in fnames)
    for fname, fh in zip(fnames, fhs):
        if count > max_samples:
            break
        keys = (k for k in fh[_sample_path_]) if filetype == 'hdf' else (k for k in fh)
        for key in keys:
            d = Sample.decode_sample_name(key)
            if d is not None:
                d['key'] = key
                d['filename'] = fname
                ref_names[d['ref_name']].append(d)
                count += 1
            if count > max_samples:
                break
    for fh in fhs:
        fh.close()

    # sort dicts so that refs are in order and within a ref, chunks are in order
    ref_names_ordered = OrderedDict()
    for ref_name in sorted(ref_names.keys()):
        sorter = lambda x: float(x['start'])
        ref_names[ref_name].sort(key=sorter)
        ref_names_ordered[ref_name] = ref_names[ref_name]

    return ref_names_ordered


def segment_limits(start, end, segment_len=20000, overlap_len=1000):
    """Generate segments of a range [0, end_point].

    :param start: startpoint of range.
    :param end: endpoint of range.
    :param segment_len: length of resultant segments.
    :param overlap_len: length of overlap between segments.
    :yields: tuples (start, end) indicating overlapping segments of
        the input range.
    """
    for n in range(start, end, segment_len):
        yield max(start, n - overlap_len), min(n + segment_len, end - 1)


def rle(array, low_mem=False):
    """Calculate a run length encoding (rle), of an input vector.

    :param array: 1D input array.
    :param low_mem: use a lower memory implementation

    returns: structured array with fields `start`, `length`, and `value`.
    """
    if len(array.shape) != 1:
        raise TypeError("Input array must be one dimensional.")
    dtype = [('length', int), ('start', int), ('value', array.dtype)]

    if not low_mem:
        pos = np.where(np.diff(array) != 0)[0]
        pos = np.concatenate(([0], pos+1, [len(array)]))
        return np.fromiter(
            ((length, start, array[start]) for (length, start) in zip(pos[1:], pos[:-1])),
            dtype, count=len(pos) - 1,
        )
    else:
        def _gen():
            start = 0
            for key, group in itertools.groupby(array):
                length = sum(1 for x in group)
                yield length, start, key
                start += length
        return np.fromiter(_gen(), dtype=dtype)


def sliding_window(a, window=3, step=1, axis=0):
    """Sliding window across an array.

    :param a: input array.
    :param window: window length.
    :param step: step length between consecutive windows.
    :param axis: axis over which to apply window.

    :yields: windows of the input array.
    """
    slicee = [slice(None)] * a.ndim
    end = 0
    for start in range(0, a.shape[axis] - window + 1, step):
        end = start + window
        slicee[axis] = slice(start, end)
        yield a[tuple(slicee)]
    # yield the remainder with the same window size
    if a.shape[axis] > end:
        start = a.shape[axis] - window
        slicee[axis] = slice(start, a.shape[axis])
        yield a[tuple(slicee)]


def mkdir_p(path, info=None):
    """Make a directory if it doesn't exist."""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            if info is not None:
                info = " {}".format(info)
            logging.warn("The path {} exists.{}".format(path, info))
            pass
        else:
            raise


class ThreadsafeIter:
    """Takes an iterator and makes it thread-safe by
    serializing call to `next` method.

    `<http://anandology.com/blog/using-iterators-and-generators/>`_
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.it)


def threadsafe_generator(f):
    """A decorator that takes a generator function and makes it
    thread-safe.

    `<http://anandology.com/blog/using-iterators-and-generators/>`_
    """
    @functools.wraps(f)
    def g(*args, **kwargs):
        return ThreadsafeIter(f(*args, **kwargs))
    return g


@threadsafe_generator
def chain_thread_safe(*gens):
    """Threadsafe version of itertools.chain"""
    gen = itertools.chain(*gens)
    while True:
        yield next(gen)


def grouper(gen, batch_size=4):
    """Group together elements of an iterable, yielding remainder without padding."""
    while True:
        batch = []
        for i in range(batch_size):
            try:
                batch.append(next(gen))
            except StopIteration:
                if len(batch) > 0:
                    yield batch
                raise StopIteration
        yield batch


def serial_gen_train_batch(xy_gen, batch_size, name=''):
    """Yield training batches."""
    count=0
    while True:
        batch = [next(xy_gen) for i in range(batch_size)]
        xs, ys = zip(*batch)
        logging.debug("Yielding {} batch {}".format(name, count))
        count += 1
        yield np.stack(xs), np.stack(ys)


@threadsafe_generator
def gen_train_batch(xy_gen, batch_size, name=''):
    yield from serial_gen_train_batch(xy_gen, batch_size, name='')


@threadsafe_generator
def yield_batches_from_hdf(h5, keys):
    """Yield training batches from HDF5.

    :param h5: `h5.File` object.
    :param keys: iterable of keys to include in the batch.
    :yields: (np.ndarray of inputs, np.ndarray of labels).
    """
    for i in keys:
        yield (h5['{}/{}'.format(_feature_batches_path_, i)][()],
               h5['{}/{}'.format(_label_batches_path_, i)][()])


class BatchQueue(object):
    def  __init__(self, batches, sparse_labels, n_classes=None, stack=3):
        """Load and queue training samples/batches from `.hdf` files.

        :param batches: tuples of (filename, hdf batch index).
        :param sparse_labels: create sparse labels.
        :param n_classes: number of classes, required if `sparse_label is False`.
        :param stack: group samples/batches by this number.

        Once initialized batches can be retrieved using batch_q._queue.get().

        """
        self.batches = batches
        self.sparse_labels = sparse_labels
        self.n_classes = n_classes
        self.stack = stack

        self.logger = copy(logging.getLogger(__package__))
        self.logger.name = 'Batcher'
        self._queue = queue.Queue(maxsize=100)
        self.stopped = threading.Event()
        self.qthread = threading.Thread(target=self._fill_queue)
        self.qthread.start()
        time.sleep(2)
        self.logger.info("Started reading batches from files.")


    def stop(self):
        self.stopped.set()
        self.logger.info("Waiting for read thread.")
        self.qthread.join(2)
        if self.qthread.is_alive:
            self.logger.critical("Read thread did not terminate.")


    def _fill_queue(self):
        with ProcessPoolExecutor(1) as executor:
            epoch = 0
            while not self.stopped.is_set():
                batch = 0
                np.random.shuffle(self.batches)
                for (fname, i) in self.batches:
                    t0 = now()
                    items = []
                    for _ in range(0, self.stack):
                        res = executor.submit(BatchQueue._generate_batch, fname, i, self.sparse_labels, self.n_classes)
                        items.append(res.result())
                    res = [np.concatenate([x[i] for x in items]) for i in range(0, 2)]
                    t1 = now()
                    self._queue.put(res)
                    self.logger.debug("Took {:5.3}s to load batch {} (epoch {})".format(t1-t0, batch, epoch))
                    batch += 1
                epoch += 1
            self.logger.info("Ended batching.")


    @staticmethod
    def _generate_batch(filename, index, sparse_labels, n_classes):
        with h5py.File(filename, 'r') as h5:
            xs = h5['{}/{}'.format(_feature_batches_path_, index)][()]
            ys = h5['{}/{}'.format(_label_batches_path_, index)][()]
        if not sparse_labels:
            ys = to_categorical(ys, num_classes=n_classes)
        return xs, ys


@threadsafe_generator
def yield_batches_from_hdfs(batches, sparse_labels=True, n_classes=None):
    """Yield batches of training features and labels indefinitely, shuffling between epochs.

    :param batches: iterable of (filename, batchname).
    :param sparse_labels: if False, labels will be one-hot encoded.
    :param n_classes: number of classes for one-hot encoding.

    :yields: (np.ndarray of inputs, np.ndarray of labels).
    
    """
    logger = copy(logging.getLogger(__package__))
    logger.name = 'BatchYd'
    queue = BatchQueue(batches, sparse_labels, n_classes)
    try:
        while True:
            yield queue._queue.get()
    except Exception as e:
        logger.critical("Exception caught why yielding batches: {}".format(e))
        queue.stop()
        raise e


def print_data_path():
    """Print data directory containing models"""
    print(resource_filename(__package__, 'data'))



from collections import Counter
from copy import copy
import functools
import inspect
import itertools
import logging
import os
from timeit import default_timer as now

import h5py
import numpy as np
import pysam

from medaka import vcf
from medaka.common import (_label_counts_path_, get_regions,
                           _label_decod_path_, decoding,
                           get_sample_index_from_files, grouper,
                           load_sample_from_hdf, load_yaml_data,
                           _model_opt_path_, mkdir_p, Sample,
                           write_sample_to_hdf, write_yaml_data,
                           yield_from_feature_files, gen_train_batch,
                           _label_batches_path_, _feature_batches_path_,
                           yield_batches_from_hdfs, _gap_,
                           _feature_decoding_path_, _feature_opt_path_)
from medaka.features import SampleGenerator


def weighted_categorical_crossentropy(weights):
    """
    A weighted version of keras.objectives.categorical_crossentropy
    @url: https://gist.github.com/wassname/ce364fddfc8a025bfab4348cf5de852d
    @author: wassname

    Variables:
        weights: numpy array of shape (C,) where C is the number of classes

    Usage:
        weights = np.array([0.5,2,10]) # Class one at 0.5, class 2 twice the normal weights, class 3 10x.
        loss = weighted_categorical_crossentropy(weights)
        model.compile(loss=loss,optimizer='adam')
    """

    from keras import backend as K
    weights = K.variable(weights)

    def loss(y_true, y_pred):
        # scale predictions so that the class probas of each sample sum to 1
        y_pred /= K.sum(y_pred, axis=-1, keepdims=True)
        # clip to prevent NaN's and Inf's
        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())
        # calc
        loss = y_true * K.log(y_pred) * weights
        loss = -K.sum(loss, -1)
        return loss

    return loss


def build_model(chunk_size, feature_len, num_classes, gru_size=128, input_dropout=0.0,
                inter_layer_dropout=0.0, recurrent_dropout=0.0):
    """Builds a bidirectional GRU model
    :param chunk_size: int, number of pileup columns in a sample.
    :param feature_len: int, number of features for each pileup column.
    :param num_classes: int, number of output class labels.
    :param gru_size: int, size of each GRU layer.
    :param input_dropout: float, fraction of the input feature-units to drop.
    :param inter_layer_dropout: float, fraction of units to drop between layers.
    :param recurrent_dropout: float, fraction of units to drop within the recurrent state.
    :returns: `keras.models.Sequential` object.
    """

    from keras.models import Sequential
    from keras.layers import Dense, GRU, Dropout
    from keras.layers.wrappers import Bidirectional

    model = Sequential()

    input_shape=(chunk_size, feature_len)
    for i in [1, 2]:
        name = 'gru{}'.format(i)
        gru = GRU(gru_size, activation='tanh', return_sequences=True, name=name,
                  dropout=input_dropout, recurrent_dropout=recurrent_dropout)
        model.add(Bidirectional(gru, input_shape=input_shape))

    if inter_layer_dropout > 0:
        model.add(Dropout(inter_layer_dropout))

    # see keras #10417 for why we specify input shape
    model.add(Dense(
        num_classes, activation='softmax', name='classify',
        input_shape=(chunk_size, 2 * feature_len)
    ))

    return model


def qscore(y_true, y_pred):
    from keras import backend as K
    error = K.cast(K.not_equal(
        K.max(y_true, axis=-1), K.cast(K.argmax(y_pred, axis=-1), K.floatx())),
        K.floatx()
    )
    error = K.sum(error) / K.sum(K.ones_like(error))
    return -10.0 * 0.434294481 * K.log(error)


def run_training(train_name, sample_gen, valid_gen, n_batch_train, n_batch_valid, feature_meta,
                 timesteps, feat_dim, model_fp=None, epochs=5000, batch_size=100, class_weight=None, n_mini_epochs=1):
    """Run training."""
    from keras.callbacks import ModelCheckpoint, CSVLogger, TensorBoard, EarlyStopping

    logger = logging.getLogger('Training')
    logger.info("Got {} batches for training, {} for validation.".format(n_batch_train, n_batch_valid))

    num_classes = len(feature_meta[_label_decod_path_])

    if model_fp is None:
        model_kwargs = { k:v.default for (k,v) in inspect.signature(build_model).parameters.items()
                         if v.default is not inspect.Parameter.empty}
    else:
        model_kwargs = load_yaml_data(model_fp, _model_opt_path_)
        assert model_kwargs is not None

    opt_str = '\n'.join(['{}: {}'.format(k,v) for k, v in model_kwargs.items()])
    logger.info('Building model with: \n{}'.format(opt_str))
    model = build_model(timesteps, feat_dim, num_classes, **model_kwargs)

    if model_fp is not None and os.path.splitext(model_fp)[-1] != '.yml':
        logger.info("Loading weights from {}".format(model_fp))
        model.load_weights(model_fp)

    logger.info("feat_dim: {}, timesteps: {}, num_classes: {}".format(feat_dim, timesteps, num_classes))
    model.summary()

    model_details = feature_meta.copy()
    model_details[_model_opt_path_] = model_kwargs
    write_yaml_data(os.path.join(train_name, 'training_config.yml'), model_details)

    opts = dict(verbose=1, save_best_only=True, mode='max')

    # define class here to avoid top-level keras import
    class ModelMetaCheckpoint(ModelCheckpoint):
        """Custom ModelCheckpoint to add medaka-specific metadata to model files"""
        def __init__(self, medaka_meta, *args, **kwargs):
            super(ModelMetaCheckpoint, self).__init__(*args, **kwargs)
            self.medaka_meta = medaka_meta

        def on_epoch_end(self, epoch, logs=None):
            super(ModelMetaCheckpoint, self).on_epoch_end(epoch, logs)
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            write_yaml_data(filepath, self.medaka_meta)

    callbacks = [
        # Best model according to training set accuracy
        ModelMetaCheckpoint(model_details, os.path.join(train_name, 'model.best.hdf5'),
                            monitor='acc', **opts),
        # Best model according to validation set accuracy
        ModelMetaCheckpoint(model_details, os.path.join(train_name, 'model.best.val.hdf5'),
                        monitor='val_acc', **opts),
        # Best model according to validation set qscore
        ModelMetaCheckpoint(model_details, os.path.join(train_name, 'model.best.val.qscore.hdf5'),
                        monitor='val_qscore', **opts),
        # Checkpoints when training set accuracy improves
        ModelMetaCheckpoint(model_details, os.path.join(train_name, 'model-acc-improvement-{epoch:02d}-{acc:.2f}.hdf5'),
                        monitor='acc', **opts),
        ModelMetaCheckpoint(model_details, os.path.join(train_name, 'model-val_acc-improvement-{epoch:02d}-{val_acc:.2f}.hdf5'),
                        monitor='val_acc', **opts),
        # Stop when no improvement, patience is number of epochs to allow no improvement
        EarlyStopping(monitor='val_loss', patience=20),
        # Log of epoch stats
        CSVLogger(os.path.join(train_name, 'training.log')),
        # Allow us to run tensorboard to see how things are going. Some
        #   features require validation data, not clear why.
        #TensorBoard(log_dir=os.path.join(train_name, 'logs'),
        #            histogram_freq=5, batch_size=100, write_graph=True,
        #            write_grads=True, write_images=True)
    ]

    if class_weight is not None:
        loss = weighted_categorical_crossentropy(class_weight)
        logger.info("Using weighted_categorical_crossentropy loss function")
    else:
        loss = 'sparse_categorical_crossentropy'
        logger.info("Using {} loss function".format(loss))

    model.compile(
       loss=loss,
       optimizer='rmsprop',
       metrics=['accuracy', qscore],
    )

    if n_mini_epochs == 1:
        logging.info("Not using mini_epochs, an epoch is a full traversal of the training data")
    else:
        logging.info("Using mini_epochs, an epoch is a traversal of 1/{} of the training data".format(n_mini_epochs))

    # fit generator
    model.fit_generator(
        sample_gen, steps_per_epoch=n_batch_train//n_mini_epochs,
        validation_data=valid_gen, validation_steps=n_batch_valid,
        max_queue_size=8, workers=8, use_multiprocessing=False,
        epochs=epochs,
        callbacks=callbacks,
        class_weight=class_weight,
    )


class VarQueue(list):

    @property
    def last_pos(self):
        if len(self) == 0:
            return None
        else:
            return self[-1].pos

    def write(self, vcf_fh):
        if len(self) > 1:
            are_dels = all(len(x.ref) == 2 for x in self)
            are_same_ref = len(set(x.chrom for x in self)) == 1
            if are_dels and are_same_ref:
                name = self[0].chrom
                pos = self[0].pos
                ref = ''.join((x.ref[0] for x in self))
                ref += self[-1].ref[-1]
                alt = ref[0]

                merged_var = vcf.Variant(name, pos, ref, alt, info=info)
                vcf_fh.write_variant(merged_var)
            else:
                raise ValueError('Cannot merge variants: {}.'.format(self))
        elif len(self) == 1:
            vcf_fh.write_variant(self[0])
        del self[:]


class VCFChunkWriter(object):
    def __init__(self, fname, chrom, start, end, reference_fasta, label_decoding):
        vcf_region_str = '{}:{}-{}'.format(chrom, start, end) #is this correct?
        self.label_decoding = label_decoding
        self.logger = logging.getLogger('VCFWriter')
        self.logger.info("Writing variants for {}".format(vcf_region_str))

        vcf_meta = ['region={}'.format(vcf_region_str)]
        self.writer = vcf.VCFWriter(fname, meta_info=vcf_meta)
        self.ref_fasta = pysam.FastaFile(reference_fasta)

    def __enter__(self):
        self.writer.__enter__()
        self.ref_fasta.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.__exit__(exc_type, exc_val, exc_tb)
        self.ref_fasta.__exit__(exc_type, exc_val, exc_tb)

    def add_chunk(self, sample, pred):
        # Write consensus alts to vcf
        cursor = 0
        var_queue = list()
        ref_seq = self.ref_fasta.fetch(sample.ref_name)
        for pos, grp in itertools.groupby(sample.positions['major']):
            end = cursor + len(list(grp))
            alt = ''.join(self.label_decoding[x] for x in pred[cursor:end]).replace(_gap_, '')
            # For simple insertions and deletions in which either
            #   the REF or one of the ALT alleles would otherwise be
            #   null/empty, the REF and ALT Strings must include the
            #   base before the event (which must be reflected in
            #   the POS field), unless the event occurs at position
            #   1 on the contig in which case it must include the
            #   base after the event
            if alt == '':
                # deletion
                if pos == 0:
                    # the "unless case"
                    ref = ref_seq[1]
                    alt = ref_seq[1]
                else:
                    # the usual case
                    pos = pos - 1
                    ref = ref_seq[pos:pos+2]
                    alt = ref_seq[pos]
            else:
                ref = ref_seq[pos]

            # Merging of variants produced by considering major.{minor} positions
            # These are of the form:
            #    X -> Y          - subs
            #    prev.X -> prev  - deletion
            #    X -> Xyy..      - insertion
            # In the second case we may need to merge variants from consecutive
            # major positions.
            if alt == ref:
                self.write(var_queue)
                var_queue = list()
            else:
                var = vcf.Variant(sample.ref_name, pos, ref, alt)
                if len(var_queue) == 0 or pos - var_queue[-1].pos == 1:
                    var_queue.append(var)
                else:
                    self.write(var_queue)
                    var_queue = [var]
            cursor = end
        self.write(var_queue)


    def write(self, var_queue):
        if len(var_queue) > 1:
            are_dels = all(len(x.ref) == 2 for x in var_queue)
            are_same_ref = len(set(x.chrom for x in var_queue)) == 1
            if are_dels and are_same_ref:
                name = var_queue[0].chrom
                pos = var_queue[0].pos
                ref = ''.join((x.ref[0] for x in var_queue))
                ref += var_queue[-1].ref[-1]
                alt = ref[0]

                merged_var = vcf.Variant(name, pos, ref, alt)
                self.writer.write_variant(merged_var)
            else:
                raise ValueError('Cannot merge variants: {}.'.format(var_queue))
        elif len(var_queue) == 1:
            self.writer.write_variant(var_queue[0])


def run_prediction(output, bam, regions, model, model_file, rle_ref, read_fraction, chunk_len, chunk_ovlp,
                   batch_size=200):
    """Inference worker."""
    logger = copy(logging.getLogger(__package__))
    logger.name = 'PWorker'

    def sample_gen():
        # chain all samples whilst dispensing with generators when done
        #   (they hold the feature vector in memory until they die)
        for region in regions:
            data_gen = SampleGenerator(
                bam, region, model_file, rle_ref, read_fraction,
                chunk_len=chunk_len, chunk_overlap=chunk_ovlp)
            yield from data_gen.samples
    batches = grouper(sample_gen(), batch_size)

    total_region_mbases = sum(r.size for r in regions) / 1e6
    logger.info("Running inference for {:.1f}M draft bases.".format(total_region_mbases))

    with h5py.File(output, 'a') as pred_h5:
        mbases_done = 0

        t0 = now()
        tlast = t0
        for data in batches:
            x_data = np.stack((x.features for x in data))
            # TODO: change to predict_on_batch?
            class_probs = model.predict(x_data, batch_size=batch_size, verbose=0)
            mbases_done += sum(x.span for x in data) / 1e6
            mbases_done = min(mbases_done, total_region_mbases)  # just to avoid funny log msg
            t1 = now()
            if t1 - tlast > 10:
                tlast = t1
                msg = '{:.1%} Done ({:.1f}/{:.1f} Mbases) in {:.1f}s'
                logger.info(msg.format(mbases_done / total_region_mbases, mbases_done, total_region_mbases, t1 - t0))

            best = np.argmax(class_probs, -1)
            for sample, prob, pred in zip(data, class_probs, best):
                # write out positions and predictions for later analysis
                if pred_h5 is not None:
                    sample_d = sample._asdict()
                    sample_d['label_probs'] = prob
                    sample_d['features'] = None  # to keep file sizes down
                    write_sample_to_hdf(Sample(**sample_d), pred_h5)

    logger.info('All done')
    return None


def predict(args):
    """Inference program."""
    os.environ["TF_CPP_MIN_LOG_LEVEL"]="2"
    from keras.models import load_model
    from keras import backend as K

    args.regions = get_regions(args.bam, region_strs=args.regions)
    logger = copy(logging.getLogger(__package__))
    logger.name = 'Predict'
    logger.info('Processing region(s): {}'.format(' '.join(str(r) for r in args.regions)))

    # write class names to output
    label_decoding = load_yaml_data(args.model, _label_decod_path_)
    write_yaml_data(args.output, {_label_decod_path_: label_decoding})

    logger.info("Setting tensorflow threads to {}.".format(args.threads))
    K.tf.logging.set_verbosity(K.tf.logging.ERROR)
    K.set_session(K.tf.Session(
        config=K.tf.ConfigProto(
            intra_op_parallelism_threads=args.threads,
            inter_op_parallelism_threads=args.threads)
    ))

    def _rebuild_model(model, chunk_len):
        time_steps = model.get_input_shape_at(0)[1]
        if chunk_len is None or time_steps != chunk_len:
            logger.info("Rebuilding model according to chunk_size: {}->{}".format(time_steps, chunk_len))
            feat_dim = model.get_input_shape_at(0)[2]
            num_classes = model.get_output_shape_at(-1)[-1]
            model = build_model(chunk_len, feat_dim, num_classes)
            logger.info("Loading weights from {}".format(args.model))
            model.load_weights(args.model)

        if logger.level == logging.DEBUG:
            model.summary()
        return model

    model = load_model(args.model, custom_objects={'qscore': qscore})

    # Split overly long regions to maximum size so as to not create
    #   massive feature matrices, then segregate those which cannot be
    #   batched.
    MAX_REGION_SIZE = int(1e6)  # 1Mb
    long_regions = []
    short_regions = []
    for region in args.regions:
        if region.size > MAX_REGION_SIZE:
            regs = region.split(MAX_REGION_SIZE, args.chunk_ovlp)
        else:
            regs = [region]
        for r in regs:
            if r.size > args.chunk_len:
                long_regions.append(r)
            else:
                short_regions.append(r)
    logger.info("Found {} long and {} short regions.".format(
        len(long_regions), len(short_regions)))

    if len(long_regions) > 0:
        logger.info("Processing long regions.")
        model = _rebuild_model(model, args.chunk_len)
        run_prediction(
            args.output, args.bam, long_regions, model, args.model, args.rle_ref, args.read_fraction,
            args.chunk_len, args.chunk_ovlp,
            batch_size=args.batch_size
        )

    # short regions must be done individually since they have differing lengths
    #   TODO: consider masking (it appears slow to apply wholesale), maybe 
    #         step down args.chunk_len by a constant factor until nothing remains.
    if len(short_regions) > 0:
        logger.info("Processing short regions")
        model = _rebuild_model(model, None)
        for region in short_regions:
            chunk_len = region.size // 2
            chunk_ovlp = chunk_len // 10 # still need overlap as features will be longer
            run_prediction(
                args.output, args.bam, [region], model, args.model, args.rle_ref, args.read_fraction,
                chunk_len, chunk_ovlp,
                batch_size=args.batch_size
            )
    logger.info("Finished processing all regions.")


def process_labels(label_counts, max_label_len=10):
    """Create map from full labels to (encoded) truncated labels."""
    logger = logging.getLogger('Labelling')

    old_labels = [k for k in label_counts.keys()]
    if type(old_labels[0]) == tuple:
        new_labels = (l[1] * decoding[l[0]].upper() for l in old_labels)
    else:
        new_labels = [l for l in old_labels]

    if max_label_len < np.inf:
        new_labels = [l[:max_label_len] for l in new_labels]

    old_to_new = dict(zip(old_labels, new_labels))
    label_decoding = list(sorted(set(new_labels)))
    label_encoding = { l: label_decoding.index(old_to_new[l]) for l in old_labels}
    logger.info("Label encoding dict is:\n{}".format('\n'.join(
        '{}: {}'.format(k, v) for k, v in label_encoding.items()
    )))

    new_counts = Counter()
    for l in old_labels:
        new_counts[label_encoding[l]] += label_counts[l]
    logger.info("New label counts {}".format(new_counts))

    return label_encoding, label_decoding, new_counts


def train(args):
    """Training program."""
    train_name = args.train_name
    mkdir_p(train_name, info='Results will be overwritten.')

    logger = logging.getLogger('Training')
    logger.debug("Loading datasets:\n{}".format('\n'.join(args.features)))

    # get feature meta data
    feature_meta = {k: load_yaml_data(args.features[0], k)
                    for k in (_feature_opt_path_, _feature_decoding_path_)}

    # get counts of labels in training samples
    label_counts = Counter()
    for f in args.features:
        label_counts.update(load_yaml_data(f, _label_counts_path_))

    logger.info("Total labels {}".format(sum(label_counts.values())))


    is_batched = True
    batches = []
    for fname in args.features:
        with h5py.File(fname, 'r') as h5:
            has_batches = _label_batches_path_ in h5 and _feature_batches_path_ in h5
            msg = 'Found batches in {}.' if has_batches else 'No batches in {}.'
            logger.info(msg.format(fname))
            batches.extend(((fname, batch) for batch in h5[_feature_batches_path_]))

    if is_batched:
        logger.info("Got {} batches.".format(len(batches)))
        # check batch size using first batch
        test_fname, test_batch = batches[0]
        with h5py.File(test_fname, 'r') as h5:
            batch_shape = h5['{}/{}'.format(_feature_batches_path_, test_batch)].shape
            label_shape = h5['{}/{}'.format(_label_batches_path_, test_batch)].shape

        logger.info("Got {} batches with feat shape {}, label shape {}".format(len(batches), batch_shape, label_shape))
        batch_size, timesteps, feat_dim = batch_shape
        if not sum(label_counts.values()) != len(batches) * timesteps:
            raise ValueError('Label counts not consistent with number of batches')

        n_batch_train = int((1 - args.validation_split) * len(batches))
        train_batches = batches[:n_batch_train]
        valid_batches = batches[n_batch_train:]
        n_batch_valid = len(valid_batches)

        if args.balanced_weights:
            sparse_labels = False
        else:
            sparse_labels = True

        gen_train = yield_batches_from_hdfs(
            train_batches, sparse_labels=sparse_labels, n_classes=len(label_counts))
        gen_valid = yield_batches_from_hdfs(
            valid_batches, sparse_labels=sparse_labels, n_classes=len(label_counts))
        label_decoding = load_yaml_data(args.features[0], _label_decod_path_)

    else:
        sample_index = get_sample_index_from_files(args.features, max_samples=args.max_samples)
        refs = [k for k in sample_index.keys()]
        logger.info("Got the following references for training:\n{}".format('\n'.join(refs)))
        n_samples = sum([len(sample_index[k]) for k in sample_index])
        logger.info("Got {} pileup chunks for training.".format(n_samples))
        # get label encoding, given max_label_len
        logger.info("Max label length: {}".format(args.max_label_len if args.max_label_len is not None else 'inf'))
        label_encoding, label_decoding, label_counts = process_labels(label_counts, max_label_len=args.max_label_len)

        # create seperate generators of x,y for training and validation
        # shuffle samples before making split so each ref represented in training
        # and validation and batches are not biased to a particular ref.
        samples = [(d['key'], d['filename']) for ref in sample_index.values() for d in ref]
        np.random.shuffle(samples)  # shuffle in place
        n_samples_train = int((1 - args.validation_split) * len(samples))
        samples_train = samples[:n_samples_train]
        samples_valid = samples[n_samples_train:]
        # all batches need to be the same size, so pad samples_train and samples_valid
        n_extra_train = args.batch_size - len(samples_train) % args.batch_size
        n_extra_valid = args.batch_size - len(samples_valid) % args.batch_size
        samples_train += [samples_train[i] for i in np.random.choice(len(samples_train), n_extra_train)]
        samples_valid += [samples_valid[i] for i in np.random.choice(len(samples_valid), n_extra_valid)]

        msg = '{} training samples padded to {}, {} validation samples padded to {}'
        logger.info(msg.format(len(set(samples_train)), len(samples_train),
                                len(set(samples_valid)), len(samples_valid)))

        # load one sample to figure out timesteps and data dim
        key, fname = samples_train[0]
        with h5py.File(fname) as h5:
            first_sample = load_sample_from_hdf(key, h5)
        timesteps = first_sample.features.shape[0]
        feat_dim = first_sample.features.shape[1]

        if not sum(label_counts.values()) // timesteps == n_samples:
            raise ValueError('Label counts not consistent with number of samples')

        gen_train_samples = yield_from_feature_files(args.features, samples=itertools.cycle(samples_train))
        gen_valid_samples = yield_from_feature_files(args.features, samples=itertools.cycle(samples_valid))
        s2xy = functools.partial(sample_to_x_y, encoding=label_encoding)
        gen_train = gen_train_batch(map(s2xy, gen_train_samples), args.batch_size, name='training')
        gen_valid = gen_train_batch(map(s2xy, gen_valid_samples), args.batch_size, name='validation')
        n_batch_train = len(samples_train) // args.batch_size
        n_batch_valid = len(samples_valid) // args.batch_size

    if args.balanced_weights:
        n_samples = sum(label_counts.values())
        n_classes = len(label_counts)
        class_weight = {k: float(n_samples)/(n_classes * count) for (k, count) in label_counts.items()}
        class_weight = np.array([class_weight[c] for c in sorted(class_weight.keys())])
    else:
        class_weight = None

    h = lambda d, i: d[i] if d is not None else 1
    logger.info("Label statistics are:\n{}".format('\n'.join(
        '{}: {} ({}, {:9.6f})'.format(i, l, label_counts[i], h(class_weight, i)) for i, l in enumerate(label_decoding)
    )))

    feature_meta[_label_decod_path_] = label_decoding
    feature_meta[_label_counts_path_] = label_counts

    run_training(train_name, gen_train, gen_valid, n_batch_train, n_batch_valid, feature_meta,
                 timesteps, feat_dim, model_fp=args.model, epochs=args.epochs, batch_size=args.batch_size,
                 class_weight=class_weight, n_mini_epochs=args.mini_epochs)





from collections import defaultdict, Counter, OrderedDict
import concurrent.futures
from copy import copy, deepcopy
import inspect
import itertools
from functools import partial
import logging
from multiprocessing import Pool
import sys
from timeit import default_timer as now

from Bio import SeqIO
import h5py
import numpy as np
import pysam

#TODO: from medaka import common
from medaka.common import (yield_compressed_pairs, Sample, lengths_to_rle, rle,
                           Region,
                           decoding, encoding, get_regions,
                           _gap_, _alphabet_, _feature_opt_path_, _feature_decoding_path_,
                           get_pairs, get_pairs_with_hp_len, seq_to_hp_lens,
                           write_yaml_data, load_yaml_data, gen_train_batch, serial_gen_train_batch,
                           _feature_batches_path_, _label_batches_path_, _label_counts_path_,
                           _label_decod_path_,)

from medaka.labels import TruthAlignment
import libmedaka


def pileup_counts(region, bam, dtype_prefixes=None, region_split=100000, workers=4):
    """Create pileup counts feature array for region.

    :param region: `Region` object
    :param bam: .bam file with alignments.
    :param dtype_prefixes: prefixes for query names which to separate counts.
        If `None` (or of length 1), counts are not split.

    :returns: pileup counts array, reference positions, insertion postions
    """
    ffi, lib = libmedaka.ffi, libmedaka.lib
    
    num_dtypes, dtypes = 1, ffi.NULL
    if isinstance(dtype_prefixes, str):
        dtype_prefixes = [dtype_prefixes]
    if dtype_prefixes is not None and len(dtype_prefixes) > 1:
        num_dtypes = len(dtype_prefixes)
        _dtypes = [ffi.new("char[]", d.encode()) for d in dtype_prefixes]
        dtypes = ffi.new("char *[]", _dtypes)

    featlen = lib.featlen

    def _process_region(reg):
        # htslib start is 1-based, Region object is 0-based
        region_str = '{}:{}-{}'.format(reg.ref_name, reg.start + 1, reg.end)

        counts = lib.calculate_pileup(
            region_str.encode(), bam.encode(), num_dtypes, dtypes
        )

        size_sizet = np.dtype(np.uintp).itemsize
        np_counts = np.frombuffer(ffi.buffer(
            counts.counts, size_sizet * counts.n_cols * featlen * num_dtypes),
            dtype=np.uintp
        ).reshape(counts.n_cols, featlen * num_dtypes).copy()

        positions = np.empty(counts.n_cols, dtype=[('major', int), ('minor', int)])
        np.copyto(positions['major'],
            np.frombuffer(ffi.buffer(
            counts.major, size_sizet * counts.n_cols),
            dtype=np.uintp
        ))
        np.copyto(positions['minor'],
            np.frombuffer(ffi.buffer(
            counts.minor, size_sizet * counts.n_cols),
            dtype=np.uintp
        ))

        lib.destroy_plp_data(counts)
        return np_counts, positions

    # split large regions, the chunks are trivially stacked
    regions = region.split(region_split)
    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
        results = list(executor.map(_process_region, regions))

    np_counts = np.concatenate([x[0] for x in results])
    positions = np.concatenate([x[1] for x in results])

    # get rid of 'first' counts row for each datatype (counts of alternative bases)
    mask = np.ones(np_counts.shape[1], dtype=bool)
    mask[[x * featlen for x in range(0, num_dtypes)]] = False
    np_counts = np_counts[:, mask]

    return np_counts, positions


class FeatureEncoder(object):
    _ref_modes_ = ['onehot', 'base_length', 'index', None]
    _norm_modes_ = ['total', 'fwd_rev', None]


    def __init__(self, ref_mode:str=None, max_hp_len:int=10,
                 log_min:int=None, normalise:str='total', with_depth:bool=False,
                 consensus_as_ref:bool=False, is_compressed:bool=True, dtypes=('',)):
        """Class to support multiple feature encodings

        :param ref_mode: str, how to represent the reference.
        :param max_hp_len: int, longest homopolymer run which can be represented, longer runs will be truncated.
        :param log_min: int, take log10 of counts/fractions and set zeros to 10**log_min.
        :param normalise: str, how to normalise the data.
        :param with_depth: bool, whether to include a feature describing the total depth.
        :param consensus_as_ref: bool, whether to use a naive max-count consensus instead of the reference.
        :param is_compressed: bool, whether to use HP compression. If false, treat as uncompressed.
        :param dtypes: iterable of str, read id prefixes of distinct data types that should be counted separately.

        """
        self.ref_mode = ref_mode
        self.consensus_as_ref = consensus_as_ref
        self.max_hp_len = max_hp_len
        self.log_min = log_min
        self.normalise = normalise
        self.feature_dtype = np.float32 if (self.normalise is not None or self.log_min is not None) else np.uint64
        self.with_depth = with_depth
        self.is_compressed = is_compressed
        self.logger = copy(logging.getLogger(__package__))
        self.logger.name = 'Feature'
        self.dtypes = dtypes

        if self.ref_mode not in self._ref_modes_:
            raise ValueError('ref_mode={} is not one of {}'.format(self.ref_mode, self._ref_modes_))
        if self.normalise not in self._norm_modes_:
            raise ValueError('normalise={} is not one of {}'.format(self.normalise, self._norm_modes_))

        opts = inspect.signature(FeatureEncoder.__init__).parameters.keys()
        opts = {k:getattr(self, k) for k in opts if k != 'self'}

        read_decoding = []
        for dtype in self.dtypes:
            # set up one-hot encoding of read run lengths for each dtype
            read_decoding += [
                (dtype,) + k
                for k in itertools.product(
                    (True, False), _alphabet_, range(1, max_hp_len + 1)
                )
            ]

            # forward and reverse gaps
            read_decoding += [(dtype, True, None, 1), (dtype, False, None, 1)]

        if self.ref_mode == 'onehot':
            ref_decoding = [('ref', b, l) for b, l in itertools.product(
                alphabet, range(1, max_hp_len + 1))]
            ref_decoding.append(('ref', _gap_, 1))  # gaps
        elif self.ref_mode == 'base_length':
            ref_decoding = ['ref_base', 'ref_length']
            self.ref_base_encoding = {b:i for i, b in enumerate(_alphabet_ + _gap_)}
        elif self.ref_mode == 'index':
            ref_decoding = ['ref_index']
        else:
            ref_decoding = []

        self.decoding = tuple(read_decoding + ref_decoding)
        if self.with_depth:
            self.decoding = self.decoding + ('depth',)
        self.encoding = OrderedDict(((a, i) for i, a in enumerate(self.decoding)))
        self.logger.debug("Creating features with: {}".format(opts))

        self.logger.debug("Label decoding is:\n{}".format('\n'.join(
            '{}: {}'.format(i, x) for i, x in enumerate(self.decoding)
        )))


    def process_ref_seq(self, ref_name, ref_fq):
        """Translate a sequence to RLE or truncated homopolymer form if required.

        :param ref_name: str, name of contig within `ref_fq`.
        :param ref_fq: path to reference fastq, or `SeqIO.index` obj.
        """
        if self.is_compressed:  # get (compressed) rle_encoded HP len from fq
            if ref_fq is None:
                raise ValueError('If homopolymers have been compressed, ref_fq must be provided')
            return get_runs_from_fastq(ref_fq, ref_name)
        elif self.max_hp_len > 1:  # get (uncompressed) hp_lens from fq
            if ref_fq is None:
                raise ValueError('If max_hp_len > 1, ref_fq must be provided')
            if isinstance(ref_fq, str):
                ref_fq = SeqIO.index(ref_fq, 'fastq')
            read = ref_fq[ref_name]
            return seq_to_hp_lens(read.seq)
        else:
            return None


    @property
    def feature_indices(self):
        """Location of feature vector components.
        
        :returns: dictionary mapping read data type and strand to feature vector
            indices (a list over all bases)

        """
        return {
            (dt, strand):
                [self.encoding[k] for k in self.encoding.keys()
                    if k[0] == dt and strand == k[1]]
                for dt, strand in itertools.product(self.dtypes, (True, False))
        }


    def bam_to_sample_c(self, reads_bam, region):
        """Converts a section of an alignment pileup (as shown
        by e.g. samtools tview) to a base frequency feature array

        :param reads_bam: (sorted indexed) bam with read alignment to reference
        :param region: `Region` object with ref_name, start and end attributes.
        :param start: starting position within reference
        :param end: ending position within reference
        :returns: `Sample` object
        """
        assert self.ref_mode is None
        assert not self.consensus_as_ref
        assert self.max_hp_len == 1
        assert self.log_min is None
        assert self.normalise == 'total' or self.normalise == 'fwd_rev' or self.normalise is None
        assert not self.with_depth
        assert not self.is_compressed

        counts, positions = pileup_counts(region, reads_bam, dtype_prefixes=self.dtypes)
        if len(counts) == 0:
            msg = 'Pileup-feature is zero-length for {} indicating no reads in this region.'.format(region)
            self.logger.warning(msg)
            return Sample(ref_name=region.ref_name, features=None,
                          labels=None, ref_seq=None,
                          positions=positions, label_probs=None)

        # find the position index for parent major position of all minor positions
        minor_inds = np.where(positions['minor'] > 0)
        major_pos_at_minor_inds = positions['major'][minor_inds]
        major_ind_at_minor_inds = np.searchsorted(positions['major'], major_pos_at_minor_inds, side='left')

        depth = np.sum(counts, axis=1)
        depth[minor_inds] = depth[major_ind_at_minor_inds]
        if self.normalise == 'total':
            # normalize counts by total depth at major position, since the
            # counts include deletions this is a count of spanning reads
            feature_array = counts / np.maximum(1, depth).reshape((-1, 1)) # max just to avoid div error
        elif self.normalise == 'fwd_rev':
            # normalize forward and reverse and by dtype
            feature_array = np.empty_like(counts, dtype=float)
            for (dt, is_rev), inds in self.feature_indices.items():
                dt_depth = np.sum(counts[:, inds], axis=1)
                dt_depth[minor_inds] = dt_depth[major_ind_at_minor_inds]
                feature_array[: , inds] = counts[:, inds] / np.maximum(1, dt_depth).reshape((-1, 1)) # max just to avoid div error
        else:
            feature_array = counts 

        feature_array = feature_array.astype(self.feature_dtype)

        sample = Sample(ref_name=region.ref_name, features=feature_array,
                        labels=None, ref_seq=None,
                        positions=positions, label_probs=None)

        self.logger.info('Processed {} (median depth {})'.format(sample.name, np.median(depth)))
        return sample


    def bam_to_sample(self, reads_bam, region, reference=None, read_fraction=None, force_py=False):
        """Converts a section of an alignment pileup (as shown
        by e.g. samtools tview) to a base frequency feature array

        :param reads_bam: (sorted indexed) bam with read alignment to reference
        :param region: `Region` object with ref_name, start and end attributes.
        :param reference: reference `.fasta`, should correspond to `bam`.
            Required only for run length encoded references and reads.
        :param read_fraction: fraction of reads to use, if `None` use all.
        :param force_py: bool, if True, force use of python code (rather than c library).
        :returns: `Sample` object
        """

        ref_rle = self.process_ref_seq(region.ref_name, reference)

        # Try to use fast c function if we can, else fall back on this function
        if not force_py and (ref_rle is None and read_fraction is None):
            try:
                return self.bam_to_sample_c(reads_bam, region)
            except Exception as e:
                self.logger.info('Could not process sample with bam_to_sample_c, using python code instead.\n({}).'.format(e))
                pass

        if self.is_compressed:
            aln_to_pairs = partial(yield_compressed_pairs, ref_rle=ref_rle)
        elif self.max_hp_len == 1:
            aln_to_pairs = get_pairs
        else:
            aln_to_pairs = partial(get_pairs_with_hp_len, ref_seq=ref_rle)

        # accumulate data in dicts
        aln_counters = defaultdict(Counter)
        ref_bases = dict()
        with pysam.AlignmentFile(reads_bam, 'rb') as bamfile:
            aln_reads = bamfile.fetch(region.ref_name, region.start, region.end)
            if read_fraction is not None:
                low, high = read_fraction
                np.random.seed((int(now()) * region.start) % 2**32)
                fraction = ((high - low) * np.random.random_sample(1) + low)[0]
                aln_reads = [a for a in aln_reads]
                n_reads = len(aln_reads)
                n_reads_to_keep = max(int(fraction * n_reads), 1)
                replace = n_reads_to_keep > n_reads
                msg = "Resampling (replace {}) from {} to {} ({:.3f}) for {}"
                self.logger.debug(msg.format(replace, n_reads, n_reads_to_keep, fraction, region))
                aln_reads = np.random.choice(aln_reads, n_reads_to_keep, replace=replace)

            start = region.start
            end = region.end
            if start is None:
                start = 0
            if end is None:
                end = float('Inf')

            for aln in aln_reads:
                # get the dtype from the prefix of the query name
                try:
                    dtype = self.dtypes[np.where([aln.query_name.startswith(dt) for dt in self.dtypes])[0][0]]
                except:
                    msg = "Skipping read {} as dtype not in {}"
                    logging.info(msg.format(aln.query_name, self.dtypes))
                    continue

                reverse = aln.is_reverse
                pairs = aln_to_pairs(aln)
                ins_count = 0
                for pair in itertools.dropwhile(lambda x: (x.rpos is None)
                                                or (x.rpos < start), pairs):
                    if ((pair.rpos == aln.reference_end - 1) or
                        (pair.rpos is not None and pair.rpos >= end)):
                        break
                    if pair.rpos is None:
                        ins_count += 1
                    else:
                        ins_count = 0
                        current_pos = pair.rpos

                    (aln_counters[(current_pos, ins_count)]
                    [self.encoding[dtype, reverse, pair.qbase, min(pair.qlen, self.max_hp_len)]]) += 1

                    ref_base = pair.rbase.upper() if pair.rbase is not None else '*'
                    ref_bases[(current_pos, ins_count)] = (ref_base, pair.rlen)


            # create feature array
            aln_cols = len(aln_counters)
            feature_len = len(self.encoding)
            feature_array = np.zeros(shape=(aln_cols, feature_len), dtype=self.feature_dtype)
            if self.log_min is not None:
                feature_array.fill(np.nan)
            ref_array = np.empty(shape=(aln_cols), dtype=[('base', int), ('run_length', int)])
            positions = np.empty(aln_cols, dtype=[('major', int),
                                                ('minor', int)])

            if aln_cols == 0:
                msg = 'Pileup-feature is zero-length for {} indicating no reads in this region.'.format(region)
                self.logger.warning(msg)
                return Sample(ref_name=region.ref_name, features=None,
                              labels=None, ref_seq=None,
                              positions=positions, label_probs=None)

            depth_array = np.empty(shape=(aln_cols), dtype=int)

            # keep track of which features are for fwd/rev reads of each dtype
            inds_by_type = self.feature_indices

            #TODO: refactor so common combinations of options can be handled as in C-function
            for i, ((pos, counts), (_, (ref_base, ref_len))) in \
                    enumerate(zip(sorted(aln_counters.items()),
                                sorted(ref_bases.items()))):
                positions[i] = pos
                ref_array[i] = (encoding[ref_base], ref_len)
                for j in counts.keys():
                    feature_array[i, j] = counts[j]

                if self.consensus_as_ref:
                    cons_i = np.argmax(feature_array[i])
                    cons_is_reverse, cons_base, cons_length = self.decoding[cons_i]
                    ref_base = cons_base if cons_base is not None else _gap_
                    ref_len = cons_length

                if positions[i]['minor'] == 0:
                    major_depth = sum(counts.values())
                    # get the depth of each fwd and rev dtype
                    major_depths_by_type = {t: sum((counts[i] for i in inds_by_type[t])) for t in inds_by_type}
                    assert sum(major_depths_by_type.values()) == major_depth

                if self.normalise is not None:
                    if self.normalise == 'total':
                        feature_array[i, :] /= max(major_depth, 1)
                    elif self.normalise == 'fwd_rev':
                        # normalize fwd and reverse seperately for each dtype
                        for dt, inds in inds_by_type.items():
                            feature_array[i, inds] /= max(major_depths_by_type[dt], 1)

                depth_array[i] = major_depth

                if self.with_depth:
                    feature_array[i, self.encoding['depth']] = depth_array[i]

                if self.log_min is not None:  # counts/proportions and depth will be normalised
                    # when we take log of probs, make it easier for network by keeping all log of
                    # probs positive. add self.log_min to any log probs so they are positive,
                    # if self.log_min is 10, we can cope with depth up to 10**9
                    feature_array[i, :] = np.log10(feature_array[i, :],
                                                   out=feature_array[i, :])
                    feature_array[i, :] += self.log_min
                    feature_array[i, :] = np.nan_to_num(feature_array[i, :], copy=False)

                if self.ref_mode == 'onehot':
                    feature_array[i, self.encoding[('ref', str(ref_base), int(ref_len))]] = 1
                elif self.ref_mode == 'base_length':
                    feature_array[i, self.encoding['ref_base']] = self.ref_base_encoding[ref_base]
                    feature_array[i, self.encoding['ref_length']] = ref_len
                elif self.ref_mode == 'index':
                    # index of count which ref would contribute to were it a read
                    feature_array[i, self.encoding['ref_index']] = self.encoding[(False, min(ref_len, self.max_hp_len), ref_base)]

            sample = Sample(ref_name=region.ref_name, features=feature_array,
                            labels=None, ref_seq=ref_array,
                            positions=positions, label_probs=None)
            self.logger.info('Processed {} (median depth {})'.format(sample.name, np.median(depth_array)))

            return sample


    def bams_to_training_samples(self, truth_bam, bam, region, reference=None, read_fraction=None):
        """Prepare training data chunks.

        :param truth_bam: .bam file of truth aligned to ref to generate labels.
        :param bam: input .bam file.
        :param region: `Region` obj.
            the reference will be parsed.
        :param reference: reference `.fasta`, should correspond to `bam`.

        :returns: tuple of `Sample` objects (one item for each input bam) for
            each chunk.

        .. note:: Chunks might be missing if `truth_bam` is provided and
            regions with multiple mappings were encountered.

        """
        ref_rle = self.process_ref_seq(region.ref_name, reference)

        # filter truth alignments to restrict ourselves to regions of the ref where the truth
        # in unambiguous
        alignments = TruthAlignment.bam_to_alignments(truth_bam, region.ref_name, start=region.start, end=region.end)
        filtered_alignments = TruthAlignment.filter_alignments(alignments, start=region.start, end=region.end)
        if len(filtered_alignments) == 0:
            self.logger.info("Filtering removed all alignments of truth to ref from {}.".format(region))

        samples = []
        for aln in filtered_alignments:
            mock_compr = self.max_hp_len > 1 and not self.is_compressed
            truth_pos, truth_labels = aln.get_positions_and_labels(ref_compr_rle=ref_rle, mock_compr=mock_compr,
                                                                   is_compressed=self.is_compressed, rle_dtype=True)
            sample = self.bam_to_sample(bam, Region(region.ref_name, aln.start, aln.end), ref_rle, read_fraction=read_fraction)
            # Create labels according to positions in pileup
            pad = (encoding[_gap_], 1) if len(truth_labels.dtype) > 0 else encoding[_gap_]
            padder = itertools.repeat(pad)
            position_to_label = defaultdict(padder.__next__,
                                            zip([tuple(p) for p in truth_pos],
                                                [a for a in truth_labels]))
            padded_labels = np.fromiter((position_to_label[tuple(p)] for p in sample.positions),
                                            dtype=truth_labels.dtype, count=len(sample.positions))

            sample = sample._asdict()
            sample['labels'] = padded_labels
            samples.append(Sample(**sample))
        return tuple(samples)


def alphabet_filter(sample_gen, alphabet=None, filter_labels=True, filter_ref_seq=True):
    """Skip chunks in which labels and/or ref_seq contain bases not in `alphabet`.

    :param sample_gen: generator of `Sample` named tuples.
    :param alphabet: set of str of allowed bases. If None, automatically generated from decoding.
    :param filter_labels: bool, whether to filter on labels.
    :param filter_ref_seq: bool, whether to filter on ref_seq.

    :yields: `Sample` named tuples.
    """
    if alphabet is None:
        alphabet = set([c for c in _alphabet_ + _gap_])
    logger = logging.getLogger('AlphaFilter')
    logger.debug("alphabet: {}".format(alphabet))

    alphabet = set([encoding[c] for c in alphabet])

    def _find_bad_bases(s, field, alphabet):
        seq_rle = getattr(s, field)
        bases = set(np.unique(seq_rle['base']))
        if not bases.issubset(alphabet):
            diff = [decoding[i] for i in bases - alphabet]
            msg = "Skipping {}:{}-{} ({} bases) due to {} {}"
            pos = s.positions
            logger.info(msg.format(s.ref_name, pos['major'][0], pos['major'][-1],
                                   len(pos), field, diff))
            return True

    for s in sample_gen:
        if filter_labels and s.labels is not None and _find_bad_bases(s, 'labels', alphabet):
            continue
        if filter_ref_seq and s.ref_seq is not None and _find_bad_bases(s, 'ref_seq', alphabet):
            continue
        yield s


class SampleGenerator(object):

    def __init__(self, bam, region, model, rle_ref=None, truth_bam=None, read_fraction=None, chunk_len=1000, chunk_overlap=200):
        """Generate chunked inference (or training) samples.

        :param bam: `.bam` containing alignments from which to generate samples.
        :param region: a `Region` for which to generate samples.
        :param model: a medaka model.
        :param truth_bam: a `.bam` containing alignment of truth sequence to
            `reference` sequence. Required only for creating training chunks.
        :param reference: reference `.fasta`, should correspond to `bam`.

        """
        self.logger = logging.getLogger("Sampler")
        self.sample_type = "training" if truth_bam is not None else "consensus"
        self.logger.info("Initializing sampler for {} or region {}.".format(self.sample_type, region))
        self.fencoder_args = load_yaml_data(model, _feature_opt_path_)
        self.fencoder = FeatureEncoder(**self.fencoder_args)

        self.bam = bam
        self.region = region
        self.model = model
        self.rle_ref = rle_ref
        self.truth_bam = truth_bam
        self.read_fraction = read_fraction
        self.chunk_len = chunk_len
        self.chunk_overlap = chunk_overlap
        self._source = None # the base data to be chunked

        #TODO: check reference has been given if model/feature encoder requires it


    def _fill_features(self):
        if self._source is None:
            t0 = now()
            if self.truth_bam is not None:
                self._source = self.fencoder.bams_to_training_samples(
                    self.truth_bam, self.bam, self.region, self.rle_ref,
                    self.read_fraction)
            else:
                self._source = self.fencoder.bam_to_sample(
                    self.bam, self.region, self.rle_ref, self.read_fraction)
                self._source = (self._source,) # wrap to be the same as above
            t1 = now()
            self.logger.info("Took {:.2f}s to make features.".format(t1-t0))

    @property
    def n_samples(self):
        """The approximate number of samples that will be yielded by `.samples`."""
        self._fill_features()
        return 1 + sum(s.size // (self.chunk_len - self.chunk_overlap) for s in self._source)


    @property
    def samples(self):
        """Iterator over chunked samples."""
        self._fill_features()
        for source in self._source:
            if source.is_empty:
                continue
            if source.size < self.chunk_len:
                msg = "Region {} ({} positions) is smaller than inference chunk length {}.".format(
                    source.name, source.size, self.chunk_len)
                self.logger.warning(msg)
                continue

            self.logger.debug(
                "Chunking pileup data into {} columns with overlap of {}.".format(
                self.chunk_len, self.chunk_overlap))
            chunks = source.chunks(chunk_len=self.chunk_len, overlap=self.chunk_overlap)
            yield from alphabet_filter(chunks)


    def training_samples(self, max_label_len):
        """Iterator of (feature, label) pairs for training."""
        self.logger.info("Maxlabellen: {}".format(max_label_len))
        if self.truth_bam is None:
            raise ValueError("Cannot iterate over training pairs when truth bam has not been given.""")
        label_encoding, label_decoding = get_label_encoding(max_label_len)
        for s in self.samples:
            if s.labels is None: # this shouldn't happen
                raise ValueError("Cannot train without labels.")
            x = s.features
            # labels can either be unicode strings or (base, length) integer tuples
            if isinstance(s.labels[0], np.unicode):
                y = np.fromiter(
                    (label_encoding[l[:min(max_label_len, len(l))]] for l in s.labels),
                    dtype=int, count=len(s.labels))
            else:
                y = np.fromiter(
                    (label_encoding[tuple((l['base'], min(max_label_len, l['run_length'])))]
                        for l in s.labels),
                    dtype=int, count=len(s.labels))
            y = y.reshape(y.shape + (1,))
            yield x, y


def get_label_encoding(max_label_len):
    """Get label encodings for a given maximum label length.

    :param max_label_len: int, maximum label length.

    :returns: (label_encoding, label_decoding_strs)
        label_encoding: {(int encoded base, int run length): int label encoding}.
        label_decoding_strs: list of str of label decodings.

    >>> get_label_encoding(2)
    ({(5, 1): 0,
      (5, 2): 1,
      (6, 1): 2,
      (6, 2): 3,
      (7, 1): 4,
      (7, 2): 5,
      (8, 1): 6,
      (8, 2): 7,
      (0, 1): 8,
      },
      ['A', 'AA', 'C', 'CC', 'G', 'GG', 'T', 'TT', '*']
    )

    """
    encoded_bases = [encoding[b] for b in _alphabet_]
    label_decoding = [(b, l) for b, l in itertools.product(
        encoded_bases, range(1, max_label_len + 1))]
    label_decoding.append((encoding[_gap_], 1))  # gaps
    label_encoding = {t: i for i, t in enumerate(label_decoding)}
    label_decoding_strs = [l * decoding[b] for (b,l) in label_decoding]
    return label_encoding, label_decoding_strs


def create_samples(args):
    raise NotImplementedError('Creation of unlabelled samples is currently disabled')


def _labelled_samples_worker(args, region):
    logger = logging.getLogger('PrepWork')
    logger.info("Processing region {}.".format(region))
    data_gen = SampleGenerator(
        args.bam, region, args.model, args.rle_ref, truth_bam = args.truth,
        read_fraction=args.read_fraction, chunk_len=args.chunk_len, chunk_overlap=args.chunk_ovlp)
    batches = serial_gen_train_batch(
        data_gen.training_samples(args.max_label_len),
        args.batch_size)
    #TODO: shuffle
    return list(batches), region, deepcopy(data_gen.fencoder_args), deepcopy(data_gen.fencoder.decoding)


def create_labelled_samples(args):
    logger = logging.getLogger('Prepare')
    regions = get_regions(args.bam, args.regions)
    reg_str = '\n'.join(['\t\t\t{}'.format(r) for r in regions])
    logger.info('Got regions:\n{}'.format(reg_str))

    labels_counter = Counter()
    fencoder_args, fencoder_decoder = None, None
    with h5py.File(args.output, 'w') as hdf:
        # TODO: this parallelism would be better in `SampleGenerator.bams_to_training_samples`
        #       since training alignments are usually chunked.
        with concurrent.futures.ProcessPoolExecutor(max_workers=args.threads) as executor:
            futures = [executor.submit(_labelled_samples_worker, args, reg) for reg in regions]
            for fut in concurrent.futures.as_completed(futures):
                if fut.exception() is None:
                    batches, region, fencoder_args, fencoder_decoder = fut.result()
                    for i, (x_batch, y_batch) in enumerate(batches):
                        logger.info("Writing batch {} for region {}".format(i, region))
                        unique, counts = np.lib.arraysetops.unique(y_batch, return_counts=True)
                        labels_counter.update(dict(zip(unique, counts)))
                        hdf['{}/{}_{}'.format(_feature_batches_path_, region, i)] = x_batch
                        hdf['{}/{}_{}'.format(_label_batches_path_, region, i)] = y_batch

    # write feature options to file, so we can later check model and features
    # are compatible and label counts, so we can weight labels by inverse counts.
    logger.info("Writing meta data to file.")
    if fencoder_args is None or fencoder_decoder is None:
        logger.critical('Feature Encoder args are (incorrectly) `None`.')

    label_encoding, label_decoding = get_label_encoding(args.max_label_len)
    to_save = {_feature_opt_path_: fencoder_args,
               _feature_decoding_path_: fencoder_decoder,
               _label_counts_path_: labels_counter,
               _label_decod_path_: label_decoding}
    for fname in (args.output, args.output + '.yml'):
        write_yaml_data(fname, to_save)


# read / reference fasta / fastq compression
_scores_ = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~'


def get_runs_from_fastq(fastq, ref_name):
    """Get run length encoding of a read.
    :param fastq: str, path to fastq file or `SeqIO.index` object.
    :param ref_name: str, name of read with fastq.
    :returns: structured `numpy.ndarray` with dtype=[('start', int), ('length', int)].
    """
    if isinstance(fastq, str):
        fastq = SeqIO.index(fastq, 'fastq')
    read = fastq[ref_name]
    lengths = read.letter_annotations['phred_quality']
    return lengths_to_rle(lengths)


def compress_seq(read):
    """Compress homopolymers within a basecall, encoding their lengths in qscores.

    :param read: `Bio.SeqRecord.SeqRecord` object
    :returns: (str read.description,
               str compressed sequence,
               str qscores,
               structured array with fields `start`, `length`, and `value`).
    """
    seq_array = np.fromiter(read.seq, dtype='U1', count=len(read.seq))
    runs = rle(seq_array, low_mem=True)
    compressed_seq = ''.join(runs['value'])
    # we can only encode up to a homopolymer length 93
    # if we want to the score decoding in pysam to correspond to counts
    # (we could get to 94 if we make score zero correspond to count 1).
    is_too_long = runs['length'] >= len(_scores_)
    if np.any(is_too_long):
        logger.warning('Some homopolymers in {} are longer than the longest supported length\n'.format(read.name))
        inds = np.where(is_too_long)
        runs['length'][inds] = len(_scores_) - 1
    compressed_scores = ''.join([ _scores_[run['length']] for run in runs])
    return read.description, compressed_seq, compressed_scores, runs


def compress(args):
    if args.output is None:
        fh = sys.stdout
    else:
        fh = open(args.output, 'w')

    formats = {'a': 'fasta', 'q': 'fastq'}
    if args.input[-1] not in formats:
        msg='Could not guess file format of {}, rename to .f(ast)a/.f(ast)q'
        raise KeyError(msg.format(args.input))

    reads = SeqIO.parse(args.input, formats[args.input[-1]])
    if args.threads > 1:
        pool = Pool(args.threads)
        compressed = pool.imap(compress_seq, reads)
    else:
        compressed = (compress_seq(r) for r in reads)

    t0 = now()
    for description, compressed_seq, compressed_scores, runs in compressed:
        fh.write('@{}\n{}\n'.format(description, compressed_seq))
        fh.write('{}\n{}\n'.format('+', compressed_scores))
    t1 = now()
    logger.info('Compressing {} took {:.3f}s.'.format(args.input, t1 - t0))

    if args.output is not None:
        fh.close()




import argparse
import logging
import os
from pkg_resources import resource_filename
import yaml

import numpy as np

from medaka.inference import train, predict
from medaka.stitch import stitch
from medaka.common import write_yaml_data
from medaka.features import create_labelled_samples, create_samples

default_model = os.path.join(resource_filename(__package__, 'data'), 'medaka_model.hdf5')

def _log_level():
    """Parser to set logging level and acquire software version/commit"""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    #parser.add_argument('--version', action='version', version=get_version())

    modify_log_level = parser.add_mutually_exclusive_group()
    modify_log_level.add_argument('--debug', action='store_const',
        dest='log_level', const=logging.DEBUG, default=logging.INFO,
        help='Verbose logging of debug information.')
    modify_log_level.add_argument('--quiet', action='store_const',
        dest='log_level', const=logging.WARNING, default=logging.INFO,
        help='Minimal logging; warnings only).')

    return parser


def _chunking_feature_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    parser.add_argument('bam', help='Input alignments.')
    parser.add_argument('--model', default=default_model, help='Model definition.')
    parser.add_argument('--batch_size', type=int, default=5, help='Inference batch size.')
    parser.add_argument('--regions', default=None, nargs='+', help='Genomic regions to analyse.')
    parser.add_argument('--chunk_len', type=int, default=10000, help='Chunk length of samples.')
    parser.add_argument('--chunk_ovlp', type=int, default=1000, help='Overlap of chunks.')
    parser.add_argument('--read_fraction', type=float, help='Fraction of reads to keep',
        nargs=2, metavar=('lower', 'upper'))
    parser.add_argument('--rle_ref', default=None, help='Encoded reference file (required only for some model types.')
    parser.add_argument('--max_label_len', type=int, default=1, help='Maximum label length.')
    return parser


def feature_gen_dispatch(args):
    if hasattr(args, 'truth'):
        create_labelled_samples(args)
    else:
        create_samples(args)


def main():
    parser = argparse.ArgumentParser('medaka',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(title='subcommands', description='valid commands', help='additional help', dest='command')
    subparsers.required = True


    # Transformation of sequence data
    #TODO: is this needed?
    #pparser = subparsers.add_parser('compress',
    #    help='Transform basecalls and draft assemblies.',
    #    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    #pparser.set_defaults(func=compress)
    #pparser.add_argument('input', help='.fasta/fastq file.')
    #pparser.add_argument('--output', default=None, help='Output file, default it stdout.')
    #pparser.add_argument('--threads', type=int, default=1, help='Number of threads for parallel execution.')

    # Creation of feature files
    fparser = subparsers.add_parser('features',
        help='Create features for inference.',
        parents=[_log_level(), _chunking_feature_args()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    fparser.set_defaults(func=feature_gen_dispatch)
    fparser.add_argument('output', help='Output features file.')
    fparser.add_argument('--truth', help='Bam of truth aligned to ref to create features for training.')
    fparser.add_argument('--threads', type=int, default=1, help='Number of threads for parallel execution.')

    # Training program
    tparser = subparsers.add_parser('train',
        help='Train a model from features.',
        parents=[_log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    tparser.set_defaults(func=train)
    tparser.add_argument('features', nargs='+', help='Path for training data.')
    tparser.add_argument('--train_name', type=str, default='keras_train', help='Name for training run.')
    tparser.add_argument('--model', help='Model definition and initial weights .hdf, or .yml with kwargs to build model.')
    tparser.add_argument('--max_label_len', type=int, default=1, help='Maximum label length.')
    tparser.add_argument('--epochs', type=int, default=5000, help='Maximum number of trainig epochs.')
    tparser.add_argument('--validation_split', type=float, default=0.2, help='Fraction of data to validate on.')
    tparser.add_argument('--batch_size', type=int, default=200, help='Training batch size.')
    tparser.add_argument('--max_samples', type=int, default=np.inf, help='Only train on max_samples.')
    tparser.add_argument('--mini_epochs', type=int, default=1, help='Reduce fraction of data per epoch by this factor')
    tparser.add_argument('--balanced_weights', action='store_true', help='Balance label weights.')

    # Consensus from bam input
    cparser = subparsers.add_parser('consensus',
        help='Run inference from a trained model and alignments.',
        parents=[_log_level(), _chunking_feature_args()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    cparser.set_defaults(func=predict)
    cparser.add_argument('output', help='Output file.')
    cparser.add_argument('--threads', type=int, default=1, help='Number of threads used by inference.')

    # Consensus from features input
    cfparser = subparsers.add_parser('consensus_from_features',
        help='Run inference from a trained model on existing features.',
        parents=[_log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    cfparser.add_argument('features', nargs='+', help='Pregenerated features (from medaka features).')
    cfparser.add_argument('--model', default=default_model, help='Model definition.')
    cfparser.add_argument('--ref_rle', default=None, help='Encoded reference file (required only for some model types.')

    # Post-processing of consensus outputs
    sparser = subparsers.add_parser('stitch',
        help='Stitch together output from medaka consensus into final output.',
        parents=[_log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    sparser.set_defaults(func=stitch)
    sparser.add_argument('inputs', nargs='+', help='Consensus .hdf files.')
    sparser.add_argument('output', help='Output .fasta.', default='consensus.fasta')
    sparser.add_argument('--regions', default=None, nargs='+', help='Limit stitching to these reference names')
    sparser.add_argument('--model_yml', help='Model yml containing label encoding, required only if consensus ended prematurely.')

    args = parser.parse_args()

    logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
    logger = logging.getLogger(__package__)
    logger.setLevel(args.log_level)

    #TODO: do common argument validation here: e.g. rle_ref being present if
    #      required by model
    args.func(args)


if __name__ == '__main__':
    main()



import os
import re
import sys
from glob import glob
import shutil
from setuptools import setup, find_packages, Extension
from setuptools import Distribution, Command
from setuptools.command.install import install
from setuptools.command.build_ext import build_ext
import subprocess

#TODO: fill in these
__pkg_name__ = 'medaka'
__author__ = 'syoung'
__description__ = 'Neural network sequence error correction.'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'

__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_name__))

verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))


dir_path = os.path.dirname(__file__)
install_requires = []
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    reqs = (
        r.split('#')[0].strip()
        for r in fh.read().splitlines() if not r.strip().startswith('#')
    )
    for req in reqs:
        if req.startswith('git+https'):
            req.split('/')[-1].split('@')[0]
        install_requires.append(req)


data_files = []
if os.environ.get('MEDAKA_BINARIES') is not None:
    exes = ['samtools', 'minimap2']
    data_files.append(
        ('exes', [
            'bincache/{}'.format(x, x) for x in exes
        ])
    )


class HTSBuild(build_ext):
    # uses the Makefile to build libhts.a, this will get done before the cffi extension
    def run(self):

        def compile_hts():
            subprocess.check_call(['make', 'libhts.a'])

        self.execute(compile_hts, [], 'Compiling htslib using Makefile')
        build_ext.run(self)


setup(
    name='medaka',
    version=__version__,
    url='https://github.com/nanoporetech/{}'.format(__pkg_name__),
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    python_requires='>=3.4.*,<3.7',
    packages=find_packages(),
    package_data={
        __pkg_name__:[os.path.join('data','*.hdf5')],
    },
    cffi_modules=["build.py:ffibuilder"],
    install_requires=install_requires,
    #place binaries as package data, below we'll copy them to standard path in dist
    data_files=data_files,
    entry_points = {
        'console_scripts': [
            '{0} = {0}.{0}:main'.format(__pkg_name__),
            'medaka_counts = {0}.medaka_counts:main'.format(__pkg_name__),
            'medaka_data_path = {0}.{1}:{2}'.format(__pkg_name__, 'common', 'print_data_path'),
        ]
    },
    scripts=['scripts/medaka_consensus', 'scripts/mini_align'],
    zip_safe=False,
    cmdclass={
        'build_ext': HTSBuild
    },
)


# Nasty hack to get binaries into bin path
class GetPaths(install):
    def run(self):
        self.distribution.install_scripts = self.install_scripts
        self.distribution.install_libbase = self.install_libbase

def get_setuptools_script_dir():
    # Run the above class just to get paths
    dist = Distribution({'cmdclass': {'install': GetPaths}})
    dist.dry_run = True
    dist.parse_config_files()
    command = dist.get_command_obj('install')
    command.ensure_finalized()
    command.run()

    print(dist.install_libbase)
    src_dir = glob(os.path.join(dist.install_libbase, 'medaka-*', 'exes'))[0]
    for exe in (os.path.join(src_dir, x) for x in os.listdir(src_dir)):
        print("Copying", os.path.basename(exe), '->', dist.install_scripts)
        shutil.copy(exe, dist.install_scripts)
    return dist.install_libbase, dist.install_scripts

if os.environ.get('MED_BINARIES') is not None:
    print("\nCopying utility binaries to your path.")
    get_setuptools_script_dir()



from copy import deepcopy
from collections import defaultdict
import itertools
from threading import Lock
import logging

from intervaltree import IntervalTree

logging.basicConfig(format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S', level=logging.INFO)
logger = logging.getLogger(__name__)


def self_return(x):
    return x


# Source: Table1 in 'The Variant Call Format Specification VCFv4.3', Table 1
# Tuples below are (number, type), where number can be:
#   A: The field has one value per alternate allele
#   R: The field has one value for each possible allele, including the reference
#   G: The field has one value for each possible genotype
#   .(dot): The number of possible values varies, is unknown or unbounded
reserved_info_fields = {'AA': (1, str), 'AC': ('A', int), 'AD': ('R', int), 'ADF': ('R', int),
                        'ADR': ('R', int), 'AF': ('A', float), 'AN': (1, int), 'BQ': (1, float),
                        'CIGAR': ('A', str), 'DB': (0, self_return), 'DP': (1, int), 'END': (1, int),
                        'H2': (0, self_return), 'H3': (0, self_return), 'MQ': (1, self_return),
                        'MQ0': (1, int), 'NS': (1, int), 'SB': ('.', self_return),
                        'SOMATIC': (0, self_return), 'VALIDATED': (0, self_return), '1000G': (0, self_return)
                        }
own_info_fields = {'SCORES': ('R', float)}
all_info_fields = reserved_info_fields.copy()
all_info_fields.update(own_info_fields)

def parse_tags_to_string(tags):
    str_tags = []
    for key, value in tags.items():
        # If key is of type 'Flag', print only key, else 'key=value'
        if value is True:
            str_tags.append(key)
        else:
            if isinstance(value, (tuple, list)):
                value = ','.join((str(x) for x in value))
            str_tags.append('{}={}'.format(key, value))
    return ';'.join(str_tags)


def parse_string_to_tags(string, splitter=','):
    tags = {}
    for field in string.split(';'):
        try:
            tag, value = field.split('=')
            if tag in all_info_fields.keys():
                _type = all_info_fields[tag][1]
                value = [_type(x) for x in value.split(splitter)]
                if len(value) == 1:
                    value = value[0]
        except ValueError:
            tag = field
            value = True

        tags[tag] = value
    return tags


class Variant(object):
    # TODO: ref/alt could be a symbolic allele "<ID>".
    # TODO: alt could contain breakends.
    # TODO: Handle genomic fields.

    def __init__(self, chrom, pos, ref, alt='.', id='.', qual='.', filter='.', info='.'):
        self.chrom = chrom
        self.pos = int(pos)
        self.ref = ref.upper()
        # self.alt should be a list/tuple of alternatives
        self.alt = alt.split(',') if isinstance(alt, str) else alt
        self.id = str(id)
        self.qual = float(qual) if qual != '.' else qual
        self.filter = filter.split(';') if ';' in filter else filter
        if isinstance(info, dict):
            self.info = info
        else:
            self.info = parse_string_to_tags(info)


    def __eq__(self, other):
        for field in ('chrom', 'pos', 'id', 'ref', 'alt', 'qual', 'filter', 'info'):
            if getattr(self, field) != getattr(other, field):
                return False
        return True


    def __ne__(self, other):
        return not self.__eq__(other)


    @property
    def info_string(self):
        return parse_tags_to_string(self.info)


    @classmethod
    def from_text(cls, line):
        chrom, pos, id, ref, alt, qual, filter, info, *others = line.split('\t')
        pos = int(pos)
        pos -= 1 # VCF is 1-based, python 0-based
        return cls(chrom, pos, ref, alt=alt, id=id, qual=qual, filter=filter, info=info)


    def add_tag(self, tag, value=None):
        self.info[tag] = value

        # Remove default value if more than one exists
        if len(self.info.keys()) > 0:
            self.info.pop('.', None)


    def get_tag(self, tag):
        return self.info[tag]


    def __repr__(self):
        attributes = {}
        for field in ('chrom', 'pos', 'ref', 'alt', 'id', 'qual', 'filter', 'info_string'):
            attributes[field] = getattr(self, field)
        return ("Variant('{chrom}', {pos}, '{ref}', alt={alt}, id={id}, qual={qual},"
                " filter={filter}, info='{info_string}')".format(**attributes))


    def deep_copy(self):
        return deepcopy(self)


class VCFWriter(object):
    def __init__(self, filename, mode='w',
                 header=('CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO'),
                 meta_info=[]
                 ):

        self.filename = filename
        self.mode = mode
        self.header = header
        self.meta = ['fileformat=VCFv4.3'] + meta_info


    def __enter__(self):
        self.handle = open(self.filename, self.mode, encoding='utf-8')
        self.handle.write('\n'.join('##' + line for line in self.meta) + '\n')
        self.handle.write('#' + '\t'.join(self.header) + '\n')
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handle.close()


    def write_variant(self, variant):
        # Some fields can be multiple
        for attribute in ('alt', 'filter'):
            value = getattr(variant, attribute)
            if isinstance(value, (tuple, list)):
                setattr(variant, attribute, ','.join(str(x) for x in value))

        # Convert info dictionary to string
        variant.info = variant.info_string

        elements = [getattr(variant, field.lower()) for field in self.header]
        # VCF POS field is 1-based
        elements[self.header.index('POS')] += 1
        line = '\t'.join([str(x) for x in elements])
        self.handle.write('{}\n'.format(line))


class VCFReader(object):
    def __init__(self, filename, cache=True):
        """Basic VCF parser.

        :param filename: .vcf file.
        :param cache: if True, all parsed variants are stored in memory for
            faster subsequent access.

        """

        self.filename = filename
        self.cache = cache
        self._indexed = False
        self._tree = None
        self._parse_lock = Lock()

        # Read both metadata and header
        self.meta = []
        self.header = None
        with open(filename, encoding='utf-8') as handle:
            for line in handle:
                line = line.replace('\n', '')
                if line.startswith('##'):
                    self.meta.append(line[2:])
                elif line.startswith('#'):
                    line = line[1:]
                    self.header = line.split('\t')
                    break


    def _parse(self):
        # just parsing the file to yield records
        last_pos = [None, None]
        with open(self.filename, encoding='utf-8') as handle:
            for index, line in enumerate(handle):
                line = line.replace('\n', '')

                # Already read meta and header in self.__init__
                if line.startswith('#'):
                    continue

                try:
                    variant = Variant.from_text(line)
                except Exception as e:
                    raise IOError(
                        'Exception while reading variant #{}.\n'
                        'Line: {}'.format(index, line)) from e

                if variant.chrom != last_pos[0]:
                    last_pos = [variant.chrom, None]
                elif last_pos[1] is not None and last_pos[1] > variant.pos:
                    raise IOError('.vcf is unsorted at index #{}.'.format(index))
                yield variant
                last_pos[1] = variant.pos


    def index(self):
        """Index the input file for faster fetches."""

        # calling this method implies caching
        self.cache = True
        if self._indexed or not self.cache:
            return

        if self._parse_lock.acquire(blocking=False):
            try:
                # clear out an incomplete parse, actually this doesn't matter since
                #    the values in the tree are set-like.
                self._tree = defaultdict(IntervalTree)
                for variant in self._parse():
                    self._tree[variant.chrom][variant.pos:variant.pos + len(variant.ref)] = variant
            except Exception:
                raise
            else:
                # record we've done a complete parse
                self._indexed = True
            finally:
                self._parse_lock.release()

        else:
            # wait for lock to be released, then return
            self._parse_lock.acquire(blocking=True)
            if not self._indexed:
                raise IOError("Waited for parsing, but parsing did not occur.")


    def fetch(self, ref_name=None, start=None, end=None, strict=True):
        """Yield all variants spanned by a region.

        :param ref_name: reference name (CHROM field of .vcf).
        :param start: inclusive start co-ordinate (0-based).
        :param end: exclusive end co-ordinate (0-based).
        :param strict: if False variants overlapping the region, but not
            contained enitrely within the region are yielded also.

        :yields: :py:class:`Variant` instances.

        """
        if start is None:
            start = float('-inf')
        if end is None:
            end = float('inf')

        if not self.cache:
            # if not using a cache, just keep re-reading the file
            for variant in self._parse():
                if not all([ref_name is None or variant.chrom == ref_name,
                            start is None or variant.pos > start,
                            end is None or variant.pos + len(variant.ref) < end]):
                    continue
                yield variant
        else:
            self.index()
            if ref_name is not None:
                results = self._tree[ref_name].search(start, end, strict=strict)
            else:
                results = itertools.chain(*(
                     x.search(start, end, strict=True)
                     for x in self._tree.values()
                ))
            # spec says .vcf is sorted, lets follow
            for interval in sorted(results):
                yield interval.data


